from collections import Counter
from collections.abc import Hashable
from typing import List

import numpy as np
from numpy.typing import NDArray

from agox.acquisitors.LCB import LowerConfidenceBoundAcquisitor
from agox.candidates.standard import StandardCandidate
from agox.databases.ABC_database import DatabaseBaseClass
from agox.models.ABC_model import ModelBaseClass
from agox.models.descriptors import SpectralGraphDescriptor, Voronoi
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
from agox.observer import Observer
from agox.writer import agox_writer


class LCBPenaltyAcquisitor(LowerConfidenceBoundAcquisitor):

    """Lower confidence bound acquisitor with a penalty term for repeated
    descriptor values.

    The acquisition function for selecting candidates is given by

    .. math::

        F(x) = E(x) - \kappa \sigma(x) + \alpha n_D(x),

    where :math:`n_D` is the number of times the descriptor value already
    exists for the candidates in the database, and :math:`\alpha` is the
    penalty scale value.

    The acquisition function for relaxing candidates is equivalent to that of
    regular LCB, as the :math:`\alpha n_D(x)` term is discontinuous.

    Parameters
    ----------
    model : ModelBaseClass
        Model to obtain energy and uncertainty predictions from.
    descriptor : DescriptorBaseClass
        Descriptor to evaluate descriptor values from.
    penalty_scale : float, optional
        Scale value for penalty (in eV), by default 1.0
    """

    def __init__(self,
                 model: ModelBaseClass,
                 descriptor: DescriptorBaseClass,
                 penalty_scale: float = 1.0,
                 **kwargs):
        super().__init__(model, **kwargs)

        if not isinstance(descriptor, (SpectralGraphDescriptor, Voronoi)):
            raise ValueError('Descriptor can currently only be graph-type descriptor')

        self.descriptor = descriptor
        self.penalty_scale = penalty_scale

        self.descriptor_value_counts = Counter()

        self.add_observer_method(self.update_descriptor_values,
                                 gets={}, sets={}, order=self.order[0],
                                 handler_identifier='database')

    def calculate_acquisition_function(self, candidates: List[StandardCandidate]) -> NDArray:
        fitness = np.zeros(len(candidates))

        for i, candidate in enumerate(candidates):
            E, sigma = self.model.predict_energy_and_uncertainty(candidate)

            descriptor_value = self._get_descriptor_value(candidate)
            count = self.descriptor_value_counts[descriptor_value]
            penalty = count * self.penalty_scale

            fitness[i] = self.acquisition_function(E, sigma) + penalty

            candidate.add_meta_information('model_energy', E)
            candidate.add_meta_information('uncertainty', sigma)
            candidate.add_meta_information('penalty', penalty)

        return fitness

    def print_information(self, candidates: List[StandardCandidate], acquisition_values: List[float]) -> None:
        if self.model.ready_state:
            for i, candidate in enumerate(candidates):
                fitness = acquisition_values[i]
                Emodel = candidate.get_meta_information('model_energy')
                sigma = candidate.get_meta_information('uncertainty')
                penalty = candidate.get_meta_information('penalty')
                self.writer('Candidate: E={:8.3f}, s={:8.3f}, p={:7.2f}, F={:8.3f}'.format(
                    Emodel, sigma, penalty, fitness))

    def attach_to_database(self, database: DatabaseBaseClass) -> None:
        if not isinstance(database, DatabaseBaseClass):
            raise TypeError(f'{database} is not a DatabaseBaseClass object')

        print(f'{self.name}: Attaching to database: {database}')
        self.attach(database)

    @agox_writer
    @Observer.observer_method
    def update_descriptor_values(self, database: DatabaseBaseClass, _) -> None:
        """Update the internal counter of descriptor values present in the
        database.
        """
        candidates = database.get_all_candidates()
        new_candidates = candidates[sum(self.descriptor_value_counts):]

        for candidate in new_candidates:
            descriptor_value = self._get_descriptor_value(candidate)
            self.descriptor_value_counts[descriptor_value] += 1

        s = 's' if len(new_candidates) != 1 else ''
        self.writer(f'Added {len(new_candidates)} structure{s} to the descriptor list')

    def _get_descriptor_value(self, candidate: StandardCandidate) -> Hashable:
        """Obtain a hashable descriptor value to store in the internal counter.

        Parameters
        ----------
        candidate : StandardCandidate
            Candidate to evaluate the descriptor for.

        Returns
        -------
        Hashable
            Hashable descriptor value.
        """
        descriptor_value = self.descriptor.get_features(candidate)

        # convert to hashable object
        if isinstance(descriptor_value, np.ndarray):
            descriptor_value = descriptor_value.data.tobytes()

        return descriptor_value



class CacheNode():
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.next = None
        self.prev = None

class Cache():
    def __init__(self, max_size=100):
        self.max_size = max_size
        self._table = {}

        self._head = None
        self._tail = None

    def get(self, key, default=None):
        if key not in self._table:
            return default

        node = self._table[key]
        
        if node is not self._tail:
            self._unlink_node(node)
            self._append_node(node)

        return node.value

    def put(self, key, value):
        if key in self._table:
            self._table[key].value = value
            self.get(key)
            return

        if len(self._table) == self.max_size:
            self._table.pop(self._head.key)
            self._remove_head_node()

        new_node = CacheNode(key=key, value=value)
        self._table[key] = new_node
        self._append_node(new_node)
        

    def _unlink_node(self, node):
        if self._head is node:
            self._head = node.next
            if node.next:
                node.next.prev = None
            return

        p, n = node.prev, node.next
        p.next, n.prev = n, p
        

    def _append_node(self, new_node):
        if not self._tail: # first time a node is cached
            self._head = self._tail = new_node
        else:
            self._tail.next = new_node
            new_node.prev = self._tail
            self._tail = self._tail.next
            

    def _remove_head_node(self):
        if not self._head:
            return

        prev = self._head
        self._head = self._head.next
        if self._head:
            self._head.prev = None
        
        del prev
    
    



from argparse import ArgumentParser
from agox.cli import parsers

def main():

    main_parser = ArgumentParser()
    subparsers = main_parser.add_subparsers(help='sub-command help')

    for add_parser in parsers:
        add_parser(subparsers)

    args = main_parser.parse_args()
    args.func(args)





from __future__ import annotations
import numpy as np
from numpy.typing import NDArray
from typing import List, Dict

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from agox.utils.metrics import get_metrics


def get_limits(
    arrays: NDArray,
    extra: float = 0.0,
    extra_min: float = None,
    extra_max: float = None,
):
    max_value = -np.inf
    min_value = np.inf

    for arr in arrays:
        max_value = max(max_value, np.max(arr))
        min_value = min(min_value, np.min(arr))

    if extra_min is not None:
        min_value -= extra_min
    if extra_max is not None:
        max_value += extra_max

    max_value += extra
    min_value -= extra

    return [min_value, max_value]


def plot_parity(
    ax: Axes,
    truths: NDArray | Dict[str, NDArray] | List[NDArray],
    predictions: NDArray | Dict[str, NDArray] | List[NDArray],
    labels=None,
    inset: bool = False,
    inset_kwargs: dict = None,
    recursive: bool = False,
    limits=None,
    metrics_used=None,
    scatter_kwargs=None,
):
    """
    Plot parity plot for a set of predictions and truths.

    Parameters
    ----------
    ax : Axes
        Matplotlib axes to plot on.
    truths : Dict[str, NDArray] | List[NDArray] | NDArray
        Dictionary with the true values.
    predictions : Dict[str, NDArray] | List[NDArray] | NDArray
        Dictionary with the predicted values.
    labels : List[str], optional
        List of keys to plot, if given only these keys will be plotted. If not,
        all pairs of keys in both truths and predictions will be plotted.
    inset : bool, optional
        Whether to plot an inset with the same data.
    inset_kwargs : dict, optional
        Kwargs to pass to the inset axes.
    metrics_used : List[str], optional
        List of metrics to display in the legend.
    scatter_kwargs : dict, optional
        Kwargs to pass to the scatter plot.
    """    

    # Scatter kwargs are created and updated with the given kwargs
    base_scatter_kwargs = {
        "edgecolor": "white",
        "linewidth": 1,
        "alpha": 1.0,
        "s": 50,
    }

    if scatter_kwargs is not None:
        base_scatter_kwargs.update(scatter_kwargs)
    
    # Limits 
    if limits is None:
        limits_extra_min = 0.1
        limits_extra_max = 0.1
    else:
        limits_extra_min = limits[0]
        limits_extra_max = limits[1]

    # Check whether we have a dictionary or a list

    if isinstance(truths, dict) and isinstance(predictions, dict):
        if labels is None:
            labels = set(truths.keys()).intersection(set(predictions.keys()))
        truths = [truths[label] for label in labels]
        predictions = [predictions[label] for label in labels]
    elif isinstance(truths, list) and isinstance(predictions, list):
        if labels is None:
            labels = [f"Pair {i}" for i in range(len(truths))]
    elif isinstance(truths, np.ndarray) and isinstance(predictions, np.ndarray):
        truths = [truths]
        predictions = [predictions]
        if labels is None:
            labels = ["Pair 1"]
    else:
        raise ValueError("Truths and predictions must be both dicts, lists or arrays.")


    for label, true, pred in zip(labels, truths, predictions):
        label_str = f"{label}"

        if metrics_used is not None:
            metrics = get_metrics(true, pred)
            for metric in metrics_used:
                label_str += f"\n{metric.upper()}: {metrics[metric]:.2f}"

        l1 = ax.scatter(
            true,
            pred,
            **base_scatter_kwargs,
            label=label_str,
        )

    # Prettyness
    calculated_limits = get_limits(
        truths + predictions, extra_max=limits_extra_max, extra_min=limits_extra_min
    )
    ax.plot(calculated_limits, calculated_limits, color="black", linestyle="--")

    # Add inset in lower right side:
    if inset:
        # Manage kwargs for the inset
        default_inset_kwargs = {
            "width": "40%",
            "height": "40%",
            "loc": "lower right",
            "bounds": [0, 5],
        }    

        default_inset_kwargs.update(inset_kwargs or {})
        inset_kwargs = default_inset_kwargs
        bounds = inset_kwargs.pop("bounds")
        ax_inset = inset_axes(ax, **inset_kwargs)

        inset_min = calculated_limits[0] + bounds[0] + limits_extra_min
        inset_max = inset_min + bounds[0] + bounds[1]
        inset_limits = [inset_min, inset_max]

        print(inset_limits)

        plot_parity(
            ax_inset,
            truths,
            predictions,
            labels=labels,
            inset=False,
            metrics_used=None,
            limits=[limits_extra_min, 0],
            recursive=True,
            scatter_kwargs=scatter_kwargs
        )
        
        ax_inset.legend(fontsize=7)
        ax_inset.set_xticklabels([])
        ax_inset.set_yticklabels([])

        ax_inset.set_xlim(inset_limits)
        ax_inset.set_ylim(inset_limits)

        # Plot the area of the inset in the main plot
        ax.add_patch(
            plt.Rectangle(
                (inset_limits[0], inset_limits[0]),
                inset_limits[1] - inset_limits[0],
                inset_limits[1] - inset_limits[0],
                fill=False,
                linestyle="--",
                color="black",
            )
        )


    if not recursive:
        ax.set_xlim(calculated_limits)
        ax.set_ylim(calculated_limits)

    ax.legend(loc="upper left")



import numpy as np
from scipy.linalg import eigvalsh
from ase.data import covalent_radii
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass

class SpectralGraphDescriptor(DescriptorBaseClass):

    name = 'SpectralGraphDescriptor'
    descriptor_type = 'global'

    def __init__(self, mode='adjacency', diagonal_mode='atomic_number', number_to_compare='all',
                 descending=False, scale_factor=1.3, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.covalent_bond_scale_factor = scale_factor
        assert mode in ['adjacency', 'laplacian']
        self.mode = mode    
        self.number_to_compare = number_to_compare

        assert diagonal_mode in ['atomic_number', 'zero']
        self.diagonal_mode = diagonal_mode

        # Ordering of eigenvalues:
        self.descending = descending

    def get_distances(self, candidate):
        distances_abs = candidate.get_all_distances(mic=True)
        numbers = candidate.get_atomic_numbers()
        r = [covalent_radii[number] for number in numbers]
        x,y = np.meshgrid(r,r)
        optimal_distances = x+y
        distances_rel = distances_abs / optimal_distances
        return distances_rel

    def get_adjacency_matrix(self, candidate):
        dist = self.get_distances(candidate)
        matrix = np.logical_and(dist > 1e-3, dist < self.covalent_bond_scale_factor).astype(int)
        if self.diagonal_mode == 'atomic_number':
            matrix += np.diag(candidate.get_atomic_numbers())
        return matrix

    def get_laplacian_matrix(self, candidate):
        A = self.get_adjacency_matrix(candidate)
        D = np.diag(np.sum(A, axis=1))
        return D - A

    def get_graph(self, candidate):
        if self.mode == 'adjacency':
            return self.get_adjacency_matrix(candidate)
        elif self.mode == 'laplacian':
            return self.get_laplacian_matrix(candidate)

    def create_features(self, candidate):
        graph = self.get_graph(candidate)

        if self.number_to_compare == 'all':
            number_to_compare = graph.shape[0]-1
        else:
            number_to_compare = self.number_to_compare        
        values = eigvalsh(graph, subset_by_index=(0, number_to_compare))

        if self.descending:
            values = np.flip(values)

        return values.reshape(1, -1)

    def get_number_of_centers(self, atoms):
        return len(atoms)

    @classmethod
    def from_atoms(cls, atoms, **kwargs):
        from agox.environments import Environment

        environment = Environment(
            template=atoms, symbols="", use_box_constraint=False, print_report=False
        )
        return cls(environment=environment, **kwargs)




import numpy as np
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass

class CenteringPostProcess(PostprocessBaseClass):

    """
    Centers a candidate object to the middle of the cell.

    Should not be used for periodic systems.
    """

    name = 'CenteringPostProcess'

    @PostprocessBaseClass.immunity_decorator
    def postprocess(self, candidate):
        """
        Centers a candidate object to the middle of the cell. 

        Parameters
        ----------
        candidate : Candidate
            Candidate object to be centered.
        
        Returns
        -------
        candidate : Candidate
            Centered candidate object.
        
        """
        com = candidate.get_center_of_mass()
        cell_middle = np.sum(candidate.get_cell(), 0) / 2
        candidate.positions = candidate.positions - com + cell_middle
        return candidate






from abc import ABC, abstractmethod
from agox.observer import Observer
from agox.writer import Writer, agox_writer


class EvaluatorBaseClass(ABC, Observer, Writer):

    """
    Base class for evaluators.

    Evaluators calculate the objective function (usually the energy).

    Parameters
    ----------
    number_to_evaluate : int, optional
        The number of candidates to evaluate before stopping. The default is 1.
    """

    def __init__(self, number_to_evaluate=1, gets={'get_key':'prioritized_candidates'}, sets={'set_key':'evaluated_candidates'}, 
        order=5, verbose=True, use_counter=True, prefix='', surname=''):
        Observer.__init__(self, gets=gets, sets=sets, order=order, surname=surname)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)
        self.number_to_evaluate = number_to_evaluate

        self.add_observer_method(self.evaluate,
                                 sets=self.sets[0], gets=self.gets[0], order=self.order[0],
                                 handler_identifier='AGOX')
    
    def __call__(self, candidate):
        return self.evaluate_candidate(candidate)
        
    @abstractmethod
    def evaluate_candidate(self, candidate):
        """
        Evaluates the given candidate.

        This function MUST append the candidates it wishes to pass to the AGOX 
        State cache to the list self.evaluated_candidates. 

        Parameters
        ----------
        candidate : AGOX candidate object.
            The candidate object to evaluate.

        Returns
        -------
        bool
            Whether the evaluation was successful. 
        """
        return successful

    @property
    @abstractmethod
    def name(self):
        pass

    @agox_writer
    @Observer.observer_method
    def evaluate(self, state):

        candidates = state.get_from_cache(self, self.get_key)
        done = False

        self.evaluated_candidates = []
        passed_evaluation_count = 0
        if self.do_check():
            while candidates and not done:
                self.writer(f'Trying candidate - remaining {len(candidates)}')
                candidate = candidates.pop(0)

                if candidate is None:
                    self.writer('Candidate was None - are your other modules working as intended?')
                    continue

                internal_state = self.evaluate_candidate(candidate)

                if internal_state:
                    self.writer('Succesful calculation of candidate.')
                    passed_evaluation_count += 1
                    self.evaluated_candidates[-1].add_meta_information('final', True)

                    if passed_evaluation_count == self.number_to_evaluate:
                        self.writer('Calculated required number of candidates.')
                        done = True
        
        state.add_to_cache(self, self.set_key, self.evaluated_candidates, mode='a')
        
    def __add__(self, other):
        return EvaluatorCollection(evaluators=[self, other])

class EvaluatorCollection(EvaluatorBaseClass): 

    name = 'EvaluatorCollection'

    def __init__(self, evaluators):
        super().__init__()
        self.evaluators = evaluators

    def evaluate_candidate(self, candidate):
        state = self.apply_evaluators(candidate)
        

    def add_evaluator(self, evaluator):
        self.evaluators.append(evaluator)

    def list_evaluators(self):
        for i, evaluator in enumerate(self.evaluators):
            print('Evaluator {}: {} - {}'.format(i, evaluator.name, evaluator))

    def apply_evaluators(self, candidate):
        for evaluator in self.evaluators:
            evaluator_state = evaluator(candidate)
            if not evaluator_state:
                return False
        return True
    
    def __add__(self, other):
        self.evaluators.append(other)
        return self



from abc import ABC, abstractmethod
from ase.calculators.singlepoint import SinglePointCalculator
#from agox import Observer, Writer, agox_writer, Module, ObserverHandler
from agox.observer import Observer, ObserverHandler
from agox.writer import Writer, agox_writer


from ase import Atoms

class DatabaseBaseClass(ABC, ObserverHandler, Observer, Writer):

    """
    Base class for all databases.

    Databases are used to store, typically evaluated, candidates. 

    Parameters
    ----------
    gets : dict
        Dictionary of get keys, e.g. {'get_key':'evaluated_candidates'}. Used to select 
        from which entry in the agox.main.State cache the database should get candidates.
    order : int
        Order of the database, by default 6
    verbose : bool
        If True, print information to the console, by default True.
    use_counter : bool
        If True the writer will not print until the Observer has finished all its 
        action in a given iteration.
    """

    def __init__(self, gets={'get_key':'evaluated_candidates'}, sets={}, order=6, verbose=True, use_counter=True, 
        prefix='', surname=''):
        Observer.__init__(self, gets=gets, sets=sets, order=order, surname=surname)
        ObserverHandler.__init__(self, handler_identifier='database', dispatch_method=self.store_in_database)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)
        self.candidates = []

        self.objects_to_assign = []

        self.add_observer_method(self.store_in_database,
                                 sets=self.sets[0], gets=self.gets[0], order=self.order[0],
                                 handler_identifier='AGOX')

    ########################################################################################
    # Required methods                                                          
    ########################################################################################

    @abstractmethod
    # def write(self, positions, energy, atom_numbers, cell, **kwargs):
    def write(self, grid):
        """
        Write stuff to database
        """

    @abstractmethod
    def store_candidate(self, candidate_object):
        pass

    @abstractmethod
    def get_all_candidates(self):
        pass

    @property
    @abstractmethod
    def name(self):
        return NotImplementedError
    
    ########################################################################################
    # Default methods
    ########################################################################################

    def __len__(self):
        return len(self.candidates) 

    @agox_writer
    @Observer.observer_method
    def store_in_database(self, state):
        
        evaluated_candidates = state.get_from_cache(self, self.get_key)
        anything_accepted = False
        for j, candidate in enumerate(evaluated_candidates):

            # Dispatch to observers only when adding the last candidate. 
            dispatch = (j+1) == len(evaluated_candidates)




            if candidate: 
                
                #self.writer('Energy {:06d}: {}'.format(self.get_iteration_counter(), candidate.get_potential_energy()), flush=True)
                self.store_candidate(candidate, accepted=True, write=True, dispatch=False)
                anything_accepted = True

            elif candidate is None:
                dummy_candidate = self.candidate_instanstiator(template=Atoms())
                dummy_candidate.set_calculator(SinglePointCalculator(dummy_candidate, energy=float('nan')))

                # This will dispatch to observers if valid data has been added but the last candidate is None. 
                self.store_candidate(candidate, accepted=False, write=True, dispatch=False)

        if anything_accepted:
            self.dispatch_to_observers(self, state)



import ray
import numpy as np

from timeit import default_timer as dt

from agox.writer import pretty_print, header_print
from agox.observer import Observer
from agox.module import Module
from agox.writer import Writer, agox_writer

from agox.utils.ray.actor import Actor

from typing import Callable, List, Dict
from time import sleep

from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy
from ray.util.placement_group import (
    placement_group,
)

from dataclasses import dataclass

@dataclass
class Task():
    """
    Task dataclass.

    Parameters
    ----------
    fn: Callable
        Function to execute.
    modules: List
        List of keys to modules to pass to the function.
    args: List
        Arguments to pass to the function.
    kwargs: Dict
        Keyword arguments to pass to the function.    
    """
    function: Callable
    modules: List
    args: List
    kwargs: Dict


class Pool(Observer, Writer, Module):

    """
    Pool class for managing a pool of actors.

    Pool of actors that can be used to execute functions in parallel.

    The pool is responsible for managing the actors and the modules that are
    attached to the actors. The pool is also responsible for updating the
    modules on the actors.

    Parameters
    ----------
    num_actors : int
        Number of actors in the pool. The number of actors should be equal to
        the number of CPU cores available.        
    """

    name = "ParallelPool"

    def __init__(self, num_actors):
        Observer.__init__(self)
        Writer.__init__(self)
        Module.__init__(self)

        # agox.modules.Module instances that are added to the pool are stored in a dict, 
        # these will be available on the actors using the keys of the dict.
        self.modules = {}

        # Start actors: Small wait time in between to avoid overloading the system.a
        self.idle_actors = []

        # Reserve a placement group for the actors.
        pg = self.reserve_placement_group(num_actors)
        self.idle_actors = []
        for actor_id in range(num_actors):
            actor = Actor.options(scheduling_strategy=PlacementGroupSchedulingStrategy(placement_group=pg)).remote(actor_id=actor_id)
            self.idle_actors.append(actor)
    
        self.number_of_actors = len(self.idle_actors)
        self.future_to_actor = {}
        self.pending_submits = []
        self.next_task_index = 0

        # Observer methods:
        self.add_observer_method(
            self.update_pool_actors,
            sets={},
            gets={},
            order=100,
            handler_identifier="AGOX",
        )
        # self.add_observer_method(
        #     self.monitoring, sets={}, gets={}, order=101, handler_identifier="AGOX"
        # )

        self.attached_handlers = []

    ##########################################################################################################
    # Using the pool - mapping out jobs.
    ##########################################################################################################

    def task_map(self, tasks: List[Task]):
        """
        Execute a list of tasks in parallel over the pool of actors.

        Parameters
        ----------
        tasks : list of agox.utils.ray.pool.Task
            List of tasks to execute.
        """
        functions = [task.function for task in tasks]
        module_keys = [task.modules for task in tasks]
        args_fn = [task.args for task in tasks]
        kwargs_fn = [task.kwargs for task in tasks]
        return self.map(functions, module_keys, args_fn, kwargs_fn)

    def map(self, functions: List[Callable], module_keys: List, args_fn: List[List], kwargs_fn: List[Dict]):
        """

        Performs the given function 'fn' in parallel over the pool of actors.

        Parameters
        ----------
        fn : function
            Function to execute on the actors.
        module_keys : list
            Keys indicating which module on the actor is passed to the function.
        args_fn : list of lists.
            Positional arguments passed to 'fn'.
        kwargs_fn : list of dicts
            Keyword-arguments passed to 'fn'.

        Returns
        -------
        list
            List of outputs such that fn(module, args[0], kwargs[0]) is the first
            element of the list.
        """

        if not isinstance(functions, list):
            functions = [functions] * len(module_keys)

        for fn, args, kwargs, module_key in zip(functions, args_fn, kwargs_fn, module_keys):
            kwargs["pool_internal_seed"] = np.random.randint(0, 10e6, size=1)
            self.submit(fn, module_key, args, kwargs)

        # Futures:
        done = False
        all_results = [None for _ in range(len(args_fn))]

        while not done:
            # Use Ray wait to get a result.
            future_ready, _ = ray.wait(list(self.future_to_actor), num_returns=1)

            # Find the actor that was working on that future:
            index, actor = self.future_to_actor.pop(future_ready[0])

            # Return the actor to the idle pool:
            self.idle_actors.append(actor)

            # Now theres a free actor we re-submit:
            if len(self.pending_submits) > 0:
                fn, module_key, args, kwargs = self.pending_submits.pop(
                    0
                )  # Pop 0 preserves the order.
                self.submit(fn, module_key, args, kwargs)

            # Because the result is ready this is 'instant' and the next job is
            # already queued.
            all_results[index] = ray.get(future_ready[0])

            if len(self.idle_actors) == self.number_of_actors:
                done = True

        # Reset the task index counter.
        self.next_task_index = 0

        return all_results

    def submit(self, fn, module_keys, args, kwargs):
        """
        Submit tasks to the worker-pool.

        Parameters
        ----------
        fn : function-handle
            Function to execute.
        module_keys : valid-key type (probably str is the best choice).
            Used to identify modules on the actor.
        args : list
            Arguments for fn
        kwargs : dict
            Key-word arguments for fn.
        """

        if len(self.idle_actors) > 0:  # Look to see if an idle actor exists.
            actor = self.idle_actors.pop()  # Take an idle actor
            future = actor.execute_function.remote(
                fn, module_keys, *args, **kwargs
            )  # Execute.
            future_key = (
                tuple(future) if isinstance(future, list) else future
            )  # Future as key
            self.future_to_actor[future_key] = (self.next_task_index, actor)  # Save
            self.next_task_index += 1  # Increment
        else:  # If no idle actors are available we save the job for later.
            # I am wondering if it would be helpful to put the args/kwargs now
            # Such that they are ready when an actor is available again.
            self.pending_submits.append((fn, module_keys, args, kwargs))

    def execute_on_actors(self, fn, module_key, args, kwargs):
        """
        Execute the function _once_ per actor with the same data.

        This could e.g. be used to set or get parameters.

        Parameters
        ----------
        fn : function
            Function to execute once pr. actor.
        module_key : list
            Modules to sue.
        args : list
            Arguments to pass to fn.
        kwargs : dict
            Key-word argument to pass to fn.

        Returns
        -------
        list
            The results of the evaluating the fn.
        """
        assert len(self.idle_actors) == self.number_of_actors
        futures = []
        for actor in self.idle_actors:
            futures += [actor.execute_function.remote(fn, module_key, *args, **kwargs)]
        return ray.get(futures)

    ##########################################################################################################
    # Managing the pool.
    ##########################################################################################################

    def get_key(self, module):
        """

        Parameters
        ----------
        module : agox.module.Module
            Module to get key for.

        Returns
        -------
        str
            The key of the module
        """
        assert isinstance(module, Module)
        return module.ray_key

    def add_module(self, module):
        """
        Parameters
        ----------
        module : AGOX module
            Module to add to the pool and its actors.

        Returns
        -------
        int
           The key that has been generateed for the module.
        """
        key = self.get_key(module)
        obj_ref = ray.put(module)
        assert len(self.idle_actors) == self.number_of_actors
        futures = [actor.add_module.remote(obj_ref, key) for actor in self.idle_actors]
        ray.get(futures)  # Block to make sure all the actors do this!
        self.modules[key] = module
        return key

    def remove_module(self, module):
        """
        Parameters
        ----------
        module : AGOX module
            Module to remove from the pool and its actors.

        """

        key = self.get_key(module)
        assert len(self.idle_actors) == self.number_of_actors
        futures = [actor.remove_module.remote(key) for actor in self.idle_actors]
        ray.get(futures)  # Block to make sure all the actors do this!
        del self.modules[key]

    @agox_writer
    @Observer.observer_method
    def update_pool_actors(self, state):
        """

        Responsible for
        1. Updating module attributes on the pool and actors.
        2. Interconnecting modules on the pool and its actors.

        The updating is handled by looking for which attributes of the supplied modules
        are marked as dynamic_attributes. E.g. for a generic AGOX Module:

            from agox.module import Module
            class GenericObserverClass(Module):

                dynamic_attributes = ['parameters']

                def __init__(self):
                    self.parameters = [1, 2, 3]
                    self.static_parameters = [2, 2, 2]

                def update_parameters(self):
                    self.parameters[0] += 1

        In this case only the 'parameters' attribute will be updated as 'static_parameters'
        are not referenced in dynamic_attributes.

        We want to internal modules on the Actors to link to each other if they use
        each other.

        If a module on the pool references another module on the pool, then they
        should also be connected on the Actors.

        Parameters
        ----------
        state : AGOX State object
            An instance of an AGOX State object.
        """
        self.update_modules()

    @agox_writer
    def monitoring(self, state):
        from ray.util.state import list_actors
        from ray.util.state import summarize_objects

        if not self.include_dashboard:
            return

        context_info = ray.get_runtime_context()
        address = context_info.gcs_address
        tab = "    "
        self.writer("--- General Information ---")
        self.writer(tab + f"Address: {address}")

        # output = subprocess.check_output(['ray', 'status', '--address', address]).decode('UTF-8').rstrip().split('\n')
        # self.writer('--- Resource usage ---')
        # for line in output:
        #     if 'CPU' in line:
        #         self.writer(tab + line)
        #     elif 'memory' in line:
        #         self.writer(tab + line)

        self.writer("--- Actor information ---")
        actor_information = list_actors(address=address)
        total_actors = len(actor_information)
        living_actors = 0
        for actor in actor_information:
            if actor["state"] == "ALIVE":
                living_actors += 1
        self.writer(tab + f"Actors: {living_actors}/{total_actors}")

        object_information = summarize_objects(address=address).get("cluster", {})
        self.writer("--- Object information ---")
        for (
            key,
            val,
        ) in object_information.items():
            if key in ["total_objects", "total_size_mb"]:
                self.writer(tab + f"{key}: {val}")

        # Memory information:
        # output = subprocess.check_output(['ray', 'memory', '--address', address, '--stats-only']).decode('UTF-8').rstrip().split('\n')
        # for l in output:
        #     print(l)

    def update_modules(self, writer=None):
        """
        Update dynamic attributes of modules on the actors.
        """
        if writer is None:
            writer = self.writer

        tab = "   "
        for i, module in enumerate(self.modules.values()):
            if not module.self_synchronizing:
                t0 = dt()
                # if len(attribute_dict) > 0:
                #     self.set_module_attributes(module, attribute_dict.keys(), attribute_dict.values())
                attribute_dict = self.synchronize_module(module, verbose=False)
                t1 = dt()
                writer(
                    tab
                    + f"{i}: {module.name} -> {len(attribute_dict)} updates in {t1-t0:04.2f} s"
                )
                for j, attribute_name in enumerate(attribute_dict.keys()):
                    writer(2 * tab + f"{j}: {attribute_name}")
            else:
                self.writer(f"{module.name} handles its own synchronization")

    def update_module_interconnections(self):
        """
        Update module interconnections on actors.

        This is done like so for each module in self.modules:
        1. Find all dynamic submodules recursively.
        2. Update the connection on the Actor.
        """
        pretty_print("Making module interconnections on actors")

        def interconnect_module(module, reference_module, setting_key):
            module.set_for_submodule(setting_key, reference_module)

        count = 0
        t0 = dt()
        tab = "   "

        # Iterate over all agox.module.Module instances in the pool.
        for module_key, module in self.modules.items():

            # Find submodules that have dynamic attributes.
            submodules = module.find_submodules(only_dynamic=True)
            for setting_key, submodule in submodules.items():
                sub_module_key = self.get_key(submodule)
                modules = [module_key, sub_module_key]
                args = [setting_key]                
                self.execute_on_actors(interconnect_module, modules, args, {})

                # Interconnect the modules on the actors.
                pretty_print(
                    tab + f"{count}: Connected {module.name} with {submodule.name}"
                )
                att_name = ".".join(setting_key)
                pretty_print(2 * tab + f"  Attribute name: {att_name}")
                count += 1
        if count == 0:
            pretty_print("No module interconnections found!")
        t1 = dt()
        pretty_print(f"Interconnecting time: {t1-t0:04.2f}")

    def print_modules(self, writer=None):
        if writer is None:
            writer = self.writer

        tab = "   "
        pretty_print("Modules in pool")
        for i, (key, module) in enumerate(self.modules.items()):
            num_dynamic = len(module.dynamic_attributes)
            report_str = tab + f"{i}: " + module.name + f" - Attrs. = {num_dynamic}"
            writer(report_str)

    def get_module_attributes(self, module, attributes):
        """
        Get one or more attributes of a module on the Actors of the pool.

        Parameters
        ----------
        module : AGOX module
            Module to get attributes from.
        attributes : list of str
            Names of the attributes to retrieve.

        Returns
        -------
        list of dicts
            A list containing the dicts that hold the requested attributes.
            The list has length equal to the number of actors and the dicts
            have length equal to the number of requested attributes.
        """

        def get_attributes(module, attributes):
            return {
                attribute: module.__dict__.get(attribute, None)
                for attribute in attributes
            }

        return self.execute_on_actors(
            get_attributes, [self.get_key(module)], [attributes], {}
        )

    def set_module_attributes(self, module, attributes, values):
        """
        Set attributes of a module on the actors of the pool.

        Parameters
        ----------
        module : AGOX module
            Module to get attributes from.
        attributes : list of str
            Names of the attributes to set.
        values : list
            Values to set for each attribute.
        """

        def set_attributes(module, attributes, values):
            for value, attribute in zip(values, attributes):
                module.__dict__[attribute] = value

        self.execute_on_actors(
            set_attributes, [self.get_key(module)], [attributes, values], {}
        )

    def synchronize_module(self, module, attributes="all", writer=None, verbose=True):
        """
        Synchronizes a single module so that its dynamic attributes are the same
        on both the main process and the actors.

        If the module is not already on the actors it will be added and subsequent
        calls will synchronize.

        Parameters
        ----------
        module : AGOX module
            Module to synchronize.
        """
        if writer is None:
            writer = self.writer
        if not verbose:
            writer = lambda x: None

        key = self.get_key(module)
        if key in self.modules.keys():
            if attributes == "all":
                attribute_dict = module.get_dynamic_attributes()
            else:
                attribute_dict = {
                    attribute: getattr(module, attribute) for attribute in attributes
                }
            writer(f"Updating attributes: {[key for key in attribute_dict.keys()]}")
            if len(attribute_dict) > 0:
                t0 = dt()
                self.set_module_attributes(
                    module, attribute_dict.keys(), attribute_dict.values()
                )
                t1 = dt()
                writer(
                    f"Updated {len(attribute_dict)} attributes on Ray Actors in {t1-t0:04.2f} s."
                )
            return attribute_dict
        else:
            self.add_module(module)
            return {}

    def reset_pool(self):
        while len(self.modules) > 0:
            key = list(self.modules.keys())[0]
            self.remove_module(self.modules[key])

    @staticmethod
    def reserve_placement_group(n_actors, n_cpus=1):    
        # Reserve a placement group for the actors.
        bundles = [{'CPU': n_cpus} for _ in range(n_actors)]

        # Create a placement group.
        pg = placement_group(bundles, strategy='PACK')
        ray.get(pg.ready())

        return pg





from .ABC_sparsifier import SparsifierBaseClass
from .CUR import CUR
from .random import RandomSparsifier
from .MBkmeans import MBkmeans

__all__ = ['SparsifierBaseClass',
            'CUR',
            'RandomSparsifier',
            'MBkmeans',
        ]



"""
Acquisitors are used to select the next candidate to evaluate among a number of
candidates, usually by sorting according to some metric.
The concept of acquisitors comes from Bayesian search algorithms.

A common acquisitor is the Lower Confidence Bound (LCB) acquisitor, which
selects according to the expression

.. math::
    LCB(x) = E(x) - \kappa \sigma(x).

Where E(x) is a prediction of the objective function at x, and :math:`\sigma(x)`
is the uncertainty of the prediction. :math:`\kappa` is a parameter that
controls the trade-off between exploration and exploitation.
This lets an algorithm choose which candidates to evaluate, but
does require the generation of multiple candidates to choose from.
"""

from .ABC_acquisitor import AcquisitonCalculatorBaseClass, AcquisitorBaseClass
from .LCB import LowerConfidenceBoundAcquisitor
from .LCB_penalty import LCBPenaltyAcquisitor
from .LCB_power import PowerLowerConfidenceBoundAcquisitor
from .meta_acquisitor import MetaInformationAcquisitor

__all__ = ['AcquisitorBaseClass',
           'LowerConfidenceBoundAcquisitor',
           'LCBPenaltyAcquisitor',
           'PowerLowerConfidenceBoundAcquisitor'
           'AcquisitonCalculatorBaseClass',
           'MetaInformationAcquisitor',
           ]



import numpy as np
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass

class ImmunizerPostprocess(PostprocessBaseClass):

    name = 'TheImmunizer'

    def __init__(self, probability=1, **kwargs):
        super().__init__(**kwargs)
        self.probability = probability

    def postprocess(self, candidate):
        if np.random.rand() < self.probability:
            candidate.set_postprocess_immunity(True)
        return candidate
        
class DeimmunizerPostprocess(PostprocessBaseClass):

    name = 'TheDeimmunizer'

    def postprocess(self, candidate):        
        candidate.set_postprocess_immunity(False)
        return candidate
        


import numpy as np
from ase.data import atomic_numbers
from ase.neighborlist import NeighborList
from matscipy.neighbours import neighbour_list as msp_nl
from scipy.spatial.distance import cdist

from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass


class ExponentialDensity(DescriptorBaseClass):

    name = "ExponentialDensity"
    descriptor_type = 'local'

    def __init__(self, indices=None, lambs=[1], rc=10.0, **kwargs):
        super().__init__(**kwargs)
        self.species = sorted(
            [atomic_numbers[i] for i in self.environment.get_species()]
        )

        self.indices = indices
        self.lambs = np.array(lambs)
        self.rc = rc

        self.n_features = len(self.species) * len(self.lambs) + 1

    def create_features(self, atoms):
        F = np.zeros((len(atoms), self.n_features))

        periodic_structure = np.any(atoms.pbc)

        if self.indices is not None:
            indices = self.indices
        else:
            indices = range(len(atoms))

        if periodic_structure:
            i1, i2, S = msp_nl("ijS", atoms, self.rc)
            D = atoms.positions[i1] - atoms.positions[i2] - S.dot(atoms.cell)
            d = np.linalg.norm(D, axis=1)

        for i in indices:
            if periodic_structure:
                args = i1 == i
                ind = i2[args]
                distances = d[args]

            else:
                ind = [j for j in range(len(atoms)) if j != i]
                pos = atoms.positions[ind]
                distances = np.linalg.norm(atoms.positions[i] - pos, axis=1)

            for c, species in enumerate(self.species):
                filter = atoms.numbers[ind] == species
                dists = distances[filter].reshape(-1, 1)
                F[i, c * len(self.lambs) : (c + 1) * len(self.lambs)] = np.sum(
                    np.exp(-dists / self.lambs) * self.gc(dists) / self.lambs, axis=0
                )

        F[indices, -1] = atoms.get_atomic_numbers()[indices]
        #        F[:, -1] = atoms.get_atomic_numbers()
        return F

    def create_feature_gradient(self, atoms):
        derivatives = np.zeros((len(atoms), len(atoms), 3, self.n_features))

        periodic_structure = np.any(atoms.pbc)
        if periodic_structure:
            i1, i2, S = msp_nl("ijS", atoms, self.rc)
            D = atoms.positions[i1] - atoms.positions[i2] - S.dot(atoms.cell)
            d = np.linalg.norm(D, axis=1)

        if self.indices is not None:
            indices = self.indices
        else:
            indices = range(len(atoms))

        for i in range(len(atoms)):
            if periodic_structure:
                args = i1 == i
                ind = i2[args]
                relative_positions = D[args]
                distances = d[args]

            else:
                ind = [j for j in range(len(atoms)) if j != i]
                positions = atoms.positions[ind]
                relative_positions = atoms.positions[i] - positions
                distances = np.linalg.norm(relative_positions, axis=1)

            for c, species in enumerate(self.species):
                filter = atoms.numbers[ind] == species
                d_vecs = relative_positions[filter]
                dists = distances[filter].reshape(-1, 1)

                for e, lamb in enumerate(self.lambs):
                    scalar_vector = (
                        np.exp(-dists / lamb)
                        * (-self.gc(dists) / lamb + self.dgc(dists))
                        / (dists * lamb)
                        * d_vecs
                    )
                    derivatives[i, i, :, c * len(self.lambs) + e] += np.sum(
                        scalar_vector, axis=0
                    )

                    feature_index = (
                        len(self.lambs) * self.species.index(atoms[i].number) + e
                    )
                    filtered_indices = np.array(ind)[filter]
                    for k in range(len(filtered_indices)):
                        derivatives[
                            filtered_indices[k], i, :, feature_index
                        ] += scalar_vector[k]

        return derivatives

    def gc(self, distances):
        filter = distances < self.rc
        values = np.zeros((len(distances), 1))
        values[filter] += 1 / 2 * np.cos(np.pi * distances[filter] / self.rc) + 1 / 2
        return values

    def dgc(self, distances):
        filter = distances < self.rc
        values = np.zeros((len(distances), 1))
        values[filter] = (
            -1 / 2 * np.pi * np.sin(np.pi * distances[filter] / self.rc) / self.rc
        )
        return values

    @classmethod
    def from_species(cls, species, **kwargs):
        from ase import Atoms

        from agox.environments import Environment

        environment = Environment(
            template=Atoms(""),
            symbols="".join(species),
            use_box_constraint=False,
            print_report=False,
        )
        return cls(environment=environment, **kwargs)

    def get_number_of_centers(self, atoms):
        return len(atoms)



from agox.writer import pretty_print, header_print
from agox.observer import Observer
from agox.utils.ray.pool_startup import get_ray_pool
from agox.utils.ray.pool import Task

from typing import List




class RayPoolUser(Observer):

    """
    Mixin to handle the parallelization of AGOX modules using Ray. 
    """
    

    def __init__(self):
        self.pool, ray_stats = get_ray_pool()
        self.cpu_count = ray_stats["CPU"]

    def pool_add_module(self, module, include_submodules=True):
        """
        Parameters
        ----------
        module : AGOX module
            The module being added to the pools modules.

        Returns
        -------
        int
            The key the pool uses for the module.
        """

        if include_submodules:
            submodules = module.find_submodules(only_dynamic=True)
            for submodule in submodules.values():
                self.pool_add_module(submodule)

        return self.pool.add_module(module)

    def pool_remove_module(self, module):
        """
        Parameters
        ----------
        module : AGOX module
            The module being added to the pools modules.
        """
        self.pool.remove_module(module)

    def pool_get_key(self, module):
        """_summary_

        Parameters
        ----------
        module : AGOX module.
            The module for which to retrieve the key.

        Returns
        -------
        int
            Key used for indexing to the module in the pool and on the actors.
        """
        return self.pool.get_key(module)

    def task_map(self, tasks: List[Task]):
        """
        Execute a list of tasks in parallel.

        Parameters
        ----------
        tasks : list of agox.utils.ray.pool.Task
            List of tasks to execute.
        """
        return self.pool.task_map(tasks)


    def pool_map(self, fn, modules, args, kwargs):
        """
        This is method that does the parallelization.

        Parameters
        ----------
        fn : function
            The function to execute in parallel.
        modules : list
            Keys for modules used by fn.
        args : list
            List of positional arguments for each call of the function.
        kwargs : list
            List of keyword arguments for the functions.

        The function 'fn' is executed once pr. set of modules, args and kwargs.
        So these lists must have the same length, and that length is equal to
        the number of parallel tasks - which does not need to equal the number
        of actors in the pool.

        As an example:

            def fn(calculator, atoms):
                atoms.set_calculator(calculator)
                return E = atoms.get_potential_energy()

            modules = [[model_key], [model_key]]
            args = [[atoms1], [atoms2]]
            kwargs = [{}, {}]

            E = pool_map(fn, modules, args, kwargs)

        Calculates the energy of two atoms objects in parallel with the calculator
        that is assumed to be a module on the Actors.

        The args and kwargs may also contain ObjectRefs obtained from ray.put.

        Returns
        -------
        list
            List of the output obtained by applying the function fn to each set
            of modules, args, kwargs.
        """
        return self.pool.map(fn, modules, args, kwargs)

    def pool_get_module_attributes(self, module, attributes):
        """

        Get one or more attributes of a module on the Actors of the pool.

        Parameters
        ----------
        module : AGOX module
            Module to get attributes from.
        attributes : list of str
            Names of the attributes to retrieve.

        Returns
        -------
        list of dicts
            A list containing the dicts that hold the requested attributes.
            The list has length equal to the number of actors and the dicts
            have length equal to the number of requested attributes.
        """
        return self.pool.get_module_attributes(module, attributes)

    def pool_set_module_attributes(self, module, attributes, values):
        """
        Set attributes of a module on the actors of the pool.

        Parameters
        ----------
        module : AGOX module
            Module to get attributes from.
        attributes : list of str
            Names of the attributes to set.
        values : list
            Values to set for each attribute.
        """
        self.pool.set_module_attributes(module, attributes, values)

    def attach(self, handler):
        """
        This method helps getting the pool attached to AGOX without needing
        to do so explicitly in a runscript.

        Parameters
        ----------
        handler : AGOX ObserverHandler instance
            The observer handler to attach to.
        """
        super().attach(handler)

        # Check if has pool:
        if not hasattr(self, "pool"):
            return

        if hash(handler) not in self.pool.attached_handlers:
            self.pool.attach(handler)
            header_print("Parallel Pool Initialization")
            self.pool.print_modules(writer=pretty_print)
            self.pool.update_module_interconnections()
            self.pool.update_modules(writer=pretty_print)
            self.pool.attached_handlers.append(hash(handler))

    def get_pool(self):
        return self.pool

    def pool_synchronize(self, attributes="all", writer=None):
        self.pool.synchronize_module(self, attributes=attributes, writer=writer)

    @classmethod
    def filter_kwargs(cls, kwargs):
        return {key: kwargs.pop(key) for key in cls.kwargs if key in kwargs}


from abc import ABC, abstractmethod
import numpy as np
from agox.observer import Observer
from agox.writer import Writer, agox_writer


class AcquisitorBaseClass(ABC, Observer, Writer):

    """
    Base-class for acquisitors.

    Parameters
    -----------
    verbose : bool
        Controls how much printing the module does.
    order : int
        Order of the observer.
    gets : dict
        Dictionary of gets, passed to the agox.observer.Observer.
    sets : dict
        Dictionary of sets, passed to the agox.observer.Observer.
    """

    def __init__(
        self,
        order=4,
        gets={"get_key": "candidates"},
        sets={"set_key": "prioritized_candidates"},
        verbose=True,
        use_counter=True,
        prefix="",
        surname="",
        skip_chance=0.0,
    ):
        Observer.__init__(self, gets=gets, sets=sets, order=order, surname=surname)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)

        self.add_observer_method(
            self.prioritize_candidates,
            sets=self.sets[0],
            gets=self.gets[0],
            order=self.order[0],
            handler_identifier="AGOX",
        )

        self.skip_chance = skip_chance

    @abstractmethod
    def calculate_acquisition_function(self, candidates):
        """
        Implements the acquisition function.

        Parameters
        -----------
        candidates : list
            List of candidates that the acquisition function will be evaluated for.

        Returns
        --------
        np.array
            Numpy array of acquisition function values.

        """
        return acquisition_values

    @agox_writer
    @Observer.observer_method
    def prioritize_candidates(self, state):
        """
        Method that is attached to the AGOX iteration loop as an observer - not intended for use outside of that loop.

        The method does the following:
        1. Gets candidates from the cache using 'get_key'.
        2. Removes 'None' from the candidate list.
        3. Calculates and sorts according to acquisition function.
        4. Adds the sorted candidates to cache with 'set_key'
        5. Prints information.

        Parameters
        -----------
        state: agox.main.State
            State object that contains the iteration data.
        """

        if self.skip_chance > 0 and np.random.rand() < self.skip_chance:
            state.add_to_cache(self, self.set_key, [], mode="a")
            return

        # Get data from the iteration data dict.
        candidate_list = state.get_from_cache(self, self.get_key)
        candidate_list = list(filter(None, candidate_list))

        # Calculate acquisition function values and sort:
        if self.do_check():
            candidate_list, acquisition_values = self.sort_according_to_acquisition_function(candidate_list)
        else:
            acquisition_values = np.zeros(len(candidate_list))
        # Add the prioritized candidates to the iteration data in append mode!
        state.add_to_cache(self, self.set_key, candidate_list, mode="a")

        self.print_information(candidate_list, acquisition_values)

    ########################################################################################
    # Default methods
    ########################################################################################

    def sort_according_to_acquisition_function(self, candidates):
        """
        Calculates acquisiton-function based on the implemeneted version calculate_acquisition_function.

        Note: Sorted so that the candidate with the LOWEST acquisition function value is first.

        Parameters
        ------------
        candidates: list
            List of candidate objects.

        Returns:
        -----------
        list
            List of candidates sorted according to increasing acquisition value (lowest first).
        np.array
            Array of acquisition function values in the same order as the sorted list, i.e. increasing.

        """
        acquisition_values = self.calculate_acquisition_function(candidates)
        sort_idx = np.argsort(acquisition_values)
        sorted_candidates = [candidates[i] for i in sort_idx]
        acquisition_values = acquisition_values[sort_idx]
        [
            candidate.add_meta_information("acquisition_value", acquisition_value)
            for candidate, acquisition_value in zip(sorted_candidates, acquisition_values)
        ]
        return sorted_candidates, acquisition_values

    def get_random_candidate(self):
        DeprecationWarning("Will be removed in the future. Please use collector.get_random_candidate")
        return self.collector.get_random_candidate()

    def print_information(self, candidates, acquisition_values):
        """
        Printing function for analysis/debugging/sanity checking.

        Parameters
        -----------
        candidates: list
            List of candidate objects.
        acquisition_values: np.array
            Acquisition function value for each candidate in 'candidates'.
        """

    def get_acquisition_calculator(self):
        """
        Creates a calculator for the acquisiton function that can be used for e.g. relaxation.

        Returns
        --------
        ASE Calculator
            ASE calculator where the energy is the acquisition function and forces are the forces of the acquisition forces.
        """
        raise NotImplementedError("'get_acqusition_calculator' is not implemented for this acquisitor")


from ase.calculators.calculator import Calculator, all_changes
from agox.module import Module


class AcquisitonCalculatorBaseClass(Calculator, Module):

    name = "AcqusitionCalculator"

    """
    Base-class for calculators that are used for acquisition functions.

    Parameters
    -----------
    model: agox.models.ABC_model.ModelBaseClass 
        A model that is used for the acquisition function.
    kwargs: dict
        Dictionary of keyword arguments passed to the Calculator-class.

    """

    def __init__(self, model, **kwargs):
        Calculator.__init__(self, **kwargs)
        Module.__init__(self)
        self.model = model

    @property
    def verbose(self):
        return self.model.verbose

    def get_model_parameters(self):
        parameters = self.model.get_model_parameters()
        return parameters

    def set_model_parameters(self, parameters):
        self.model.set_model_parameters(parameters)

    def set_verbosity(self, verbose):
        self.model.set_verbosity(verbose)

    def get_iteration_number(self):
        if hasattr(self, "get_iteration_counter"):
            return self.get_iteration_counter()
        else:
            return self.iteration

    @property
    def ready_state(self):
        return self.model.ready_state



import pytest
import numpy as np
from agox.environments.environment import Environment
from ase import Atoms
from ase.build import fcc100
from agox.generators import RandomGenerator, RattleGenerator
from agox.samplers.fixed_sampler import FixedSampler

def nonperiodic_environment(symbols):
    template = Atoms("", cell=np.eye(3) * 25, pbc=False)
    confinement_corner = np.array([2.5, 2.5, 2.5])
    confinement_cell = np.eye(3) * 20
    environment = Environment(
        template=template,
        symbols=symbols,
        confinement_corner=confinement_corner,
        confinement_cell=confinement_cell,
    )
    return environment

def slab_environment(symbols):
    template = fcc100('Al', size=(3, 3, 2), vacuum=10.0, orthogonal=True, periodic=True)
    template.positions[:, 2] -= template.positions[:, 2].min()

    confinement_corner = [0, 0, template.positions[:, 2].max()]
    confinement_cell = template.cell.copy()
    confinement_cell[2, 2] = 4

    environment = Environment(
        template=template, 
        symbols=symbols,
        confinement_corner=confinement_corner,
        confinement_cell=confinement_cell
    )
    return environment

non_periodic_environment_params = [
    (nonperiodic_environment, ['B12']),
    (nonperiodic_environment, ['B12N12'])
]
non_periodic_environment_param_ids = ['nonperiodic_B12', 'nonperiodic_B12N12']

@pytest.fixture(params=non_periodic_environment_params, 
                ids=non_periodic_environment_param_ids,
                scope='module')
def test_environment_non_periodic(request):
    return request.param[0](*request.param[1])

slab_environment_params = [
    (slab_environment, ['Al9']),
    (slab_environment, ['Al5O4'])
]
slab_environment_param_ids = ['slab_Al9', 'slab_Al5O4']

@pytest.fixture(params=slab_environment_params, 
                ids=slab_environment_param_ids,
                scope='module')
def test_environment_slab(request):
    return request.param[0](*request.param[1])


all_environment_params = slab_environment_params + non_periodic_environment_params
all_environment_param_ids = slab_environment_param_ids + non_periodic_environment_param_ids

@pytest.fixture(params=all_environment_params,
    ids=all_environment_param_ids,
    scope='module')
def test_environment(request):
    return request.param[0](*request.param[1])

@pytest.fixture(scope='module')
def random_generator():

    def make_generator(environment):
        generator = RandomGenerator(**environment.get_confinement())
        return generator
    return make_generator

@pytest.fixture(scope='module')
def random_candidate(random_generator, test_environment):
    generator = random_generator(test_environment)
    return generator(None, test_environment)[0]

@pytest.fixture(scope='module')
def sampler(random_candidate):
    return FixedSampler(random_candidate)




from argparse import ArgumentParser
from agox.databases.database import Database
from agox.models.descriptors.voronoi import Voronoi
from ase.io import write, read
import os

class Analysis:

    def __init__(self, descriptor, directories, sample_size, save_data = True, force_reload = False):
        self.descriptor = descriptor
        self.directories = directories
        self.sample_size = sample_size
        self.analysed_directories = []

        self.save_data = save_data
        self.force_reload = force_reload
    
        self.structures = []

    def analyse_directories(self):        
        for idx, directory in enumerate(self.directories):
            self.add_directory(directory, force_reload=self.force_reload)

    def add_directory(self, directory, force_reload = False):

        if directory in self.analysed_directories:
            print('Directory: {} - already been loaded so skipping it'.format(directory))
        
        cwd = os.getcwd()
        full_path = os.path.join(cwd, directory)
        if full_path[-1] != '/':
            full_path += '/'
        data_path = full_path + 'GS_data/'

        if os.path.exists(data_path) and not force_reload:
            print('{} has already been analysed.'.format(full_path))
            self.structures += read(data_path + 'unique.traj', index = ':')
            if self.descriptor.template is None:
                template = read(data_path + 'template.traj', index = '0')
                self.descriptor.template = template

        else:                    
            # Find database files in this directory:
            structures = []     
            print('Loading files from directory: {}'.format(directory))
            for f in os.listdir(os.path.join(cwd, directory)):
                if f[-3:] == '.db':
                    db = Database(os.path.join(cwd, directory, f))
                    db.restore_to_memory()
                    structures += db.get_all_candidates()
                    if self.descriptor.template is None:
                        template = structures[0].get_template()
                        self.descriptor.template = template

            unique = self.sort_structures(structures)
            self.structures += unique

            if self.save_data:
                if not os.path.exists(data_path):
                    os.mkdir(data_path)
                write(data_path + 'unique.traj', unique)
                write(data_path + 'template.traj', self.descriptor.template)

        self.analysed_directories.append(directory)

    def sort_structures(self, structures):
        structures = sorted(structures, key = lambda x: x.get_potential_energy())
        print(f'Total number of structures: {len(structures)}')
        unique = []
        eigen_spectrum = []
        for i, structure in enumerate(structures):
            eigen_values = self.descriptor.create_features(structure)
            if eigen_values not in eigen_spectrum:
                unique.append(structure)
                eigen_spectrum.append(eigen_values)
                print(i, eigen_values)
                if len(unique) == self.sample_size:
                    break
        print(f'Number of unique structures: {len(unique)}')
        return unique


def analysis():
    parser = ArgumentParser()
    parser.add_argument('-d', '--directories', nargs='+', type=str, default='')  # List - Directories
    parser.add_argument('-t', '--trajectories', nargs='+', type=str, default='')
    parser.add_argument('-fr', '--force_reload', action='store_true')
    parser.add_argument('-sd', '--save_data', action='store_false')

    parser.add_argument('-ss', '--sample_size', default=10, type = int)
    parser.add_argument('-indices', '--indices', default=None)
    parser.add_argument('-template', '--template', default=None)
    parser.add_argument('-cbf', '--covalent_bond_scale_factor', type=float, default=1.3)
    parser.add_argument('-angle', '--angle_from_central_atom', type=float, default=20.)
    parser.add_argument('-n_points', '--number_of_points', type=int, default=8)

    args = parser.parse_args()
    directories = args.directories
    trajectories = args.trajectories

    save_data = args.save_data
    force_reload = args.force_reload

    sample_size = args.sample_size

    covalent_bond_scale_factor = args.covalent_bond_scale_factor
    n_points = args.number_of_points
    angle = args.angle_from_central_atom
    template = args.template
    indices = args.indices

    descriptor = Voronoi(indices = indices, template = template, covalent_bond_scale_factor = covalent_bond_scale_factor, n_points = n_points, angle_from_central_atom=angle, environment=None)

    A = Analysis(descriptor, directories, sample_size, save_data=save_data, force_reload=force_reload)
    if len(directories) != 0:
        A.analyse_directories()
    
    if len(trajectories) != 0:
        print('Loading structures from trajectories')
        structures = []
        for trajectory in trajectories:
            structures += read(trajectory, index = ':')

        A.structures += structures

    if len(directories) == 0 and len(trajectories) == 0:
        print('No directories or trajectories have been provided.')
        return
    
    print('\nConstructing combined sample from all directories and trajectories.\n')
    A.sample_size = 10000
    combined_unique = A.sort_structures(A.structures)
    write('combined_sample.traj', combined_unique)

if __name__ == '__main__':
    analysis()



import numpy as np

class ConstantPrior:

    def __init__(self, value=1):
        self.value = value

    def predict_energy(self, candidate):
        return self.value

    def predict_forcces(self, candidate):
        return np.zeros((len(candidate), 3))


import itertools
from typing import Any, Dict, List, Optional, Union

import ase.data
import ase.data.colors
import numpy as np
from ase.atoms import Atoms
from ase.constraints import FixAtoms
from matplotlib.axes import Axes
from matplotlib.collections import PatchCollection
from matplotlib.patches import Circle, PathPatch
from matplotlib.path import Path

from agox.utils.plot.colors import Colors, ColorType
from agox.utils.plot.utils import plane_to_indices


def plot_atoms(ax: Axes,
               atoms: Atoms,
               *,
               colors: Optional[Union[Colors, List[ColorType]]] = None,
               plane: str = 'xy+',
               radius_factor: float = 1.0,
               repeat: int = 0,
               plot_constraint: bool = False,
               patch_kwargs: Optional[Dict[str, Any]] = None,
               repeat_patch_kwargs: Optional[Dict[str, Any]] = None) -> PatchCollection:
    """Plot a 2D projection of an `Atoms` object into an `Axes` object.

    Parameters
    ----------
    ax : Axes
        Axes to plot the atoms in.
    atoms : Atoms
        Atoms to plot.
    colors: Optional[Union[Colors, List[ColorType]]], optional
        List of colors for every atom. If None, use the jmol colors. The
        `Colors` class can be used and provided to easily generate a list of
        colors.
    plane : str, optional
        Plane to project the atoms on. See `plane_to_indices` for information
        on valid plane strings, by default 'xy+'.
    radius_factor : float, optional
        Factor to multiply the atomic covalent radii with, by default 1.0.
    repeat : int, optional
        Number of times to repeat the Atoms object in dimensions with periodic
        boundary conditions, by default 0. Repeated copies are not included in
        autoscaling. Note that if none of the two projection axes have periodic
        boundary conditions, this parameter is effectively ignored.
    plot_constraint : bool, optional
        Whether to draw crosses on top of the atoms that are fixed by a
        `FixAtoms` constraint, similar to the ASE GUI, by default False.
    patch_kwargs : Optional[Dict[str, Any]], optional
        Keyword arguments forwarded to the `Circle` constructor for each atom,
        by default None.
    repeat_patch_kwargs : Optional[Dict[str, Any]], optional
        Keyword arguments additionally forwarded to the `Circle` constructor
        for each repeated atom, by default None.

    Returns
    -------
    PatchCollection
        Collection containing patches that make up the Atoms object.
    """

    if len(atoms) == 0:
        return PatchCollection([])

    if colors is None:
        colors = Colors(atoms)

    if patch_kwargs is None:
        patch_kwargs = {}
    if repeat_patch_kwargs is None:
        repeat_patch_kwargs = {}

    d0, d1, d2, kv = plane_to_indices(plane)

    radii = radius_factor * ase.data.covalent_radii[atoms.numbers]

    fixed = np.zeros(len(atoms), dtype=bool)
    if plot_constraint:
        for constraint in atoms.constraints:
            if isinstance(constraint, FixAtoms):
                fixed[constraint.index] = True

    xy_min = np.min(atoms.positions[:, [d0, d1]] - radii[:, np.newaxis], axis=0)
    xy_max = np.max(atoms.positions[:, [d0, d1]] + radii[:, np.newaxis], axis=0)

    if repeat > 0:
        atoms_ = atoms.copy()
        pbc = atoms.get_pbc()

        repeat_range_d0 = range(-repeat, repeat + 1) if pbc[d0] else [0]
        repeat_range_d1 = range(-repeat, repeat + 1) if pbc[d1] else [0]

        for x, y in itertools.product(repeat_range_d0, repeat_range_d1):
            if x == y == 0:
                continue
            offset_uc = np.zeros(3)
            offset_uc[[d0, d1]] = [x, y]
            offset = atoms.cell.cartesian_positions(offset_uc)
            repeat_atoms = atoms.copy()
            repeat_atoms.translate(offset)
            atoms_ += repeat_atoms
    else:
        atoms_ = atoms

    n_atoms = len(atoms)
    patches: List[Circle] = []

    for i in np.argsort(kv * atoms_.positions[:, d2]):
        patch_kwargs_ = {**patch_kwargs} if i < n_atoms else {**patch_kwargs, **repeat_patch_kwargs}

        position = atoms_[i].position
        radius = radii[i % n_atoms]
        color = colors[i % n_atoms]

        circle = Circle(position[[d0, d1]], radius=radius, ec='k', fc=color, **patch_kwargs_)

        patches.append(circle)

        if fixed[i % n_atoms]:
            path_patch_kwargs = {k: v for k, v in patch_kwargs_.items() if k not in ['fc', 'facecolor']}

            delta = radius * 0.5 * np.array([np.sqrt(2), np.sqrt(2)])

            p1 = position[[d0, d1]] + delta * [-1, 1]
            p2 = position[[d0, d1]] + delta * [1, -1]
            p3 = position[[d0, d1]] + delta * [-1, -1]
            p4 = position[[d0, d1]] + delta * [1, 1]

            verts = [p1, p2, p3, p4]
            codes = [Path.MOVETO, Path.LINETO, Path.MOVETO, Path.LINETO]
            path = Path(verts, codes)
            path_patch = PathPatch(path, **path_patch_kwargs)

            patches.append(path_patch)

    p = PatchCollection(patches, match_original=True)

    ax.add_collection(p, autolim=False)
    ax.update_datalim((xy_min, xy_max))
    ax.set_aspect('equal')
    ax.tick_params(bottom=False, top=False, left=False, right=False,
                   labelbottom=False, labeltop=False,
                   labelleft=False, labelright=False)
    ax.autoscale_view()

    return p



import ray
import numpy as np
from agox.collectors.ABC_collector import CollectorBaseClass
from agox.utils.ray import RayPoolUser, Task


def remote_generate(generator, sampler, environment):
    return generator(sampler, environment)

class ParallelCollector(CollectorBaseClass, RayPoolUser):

    """
    This collector uses a pool of actors to generate candidates in parallel.

    Parameters
    ----------
    num_candidates : dict or list
        If dict, it must have the form {0: [], 500: []} where the keys are the iteration numbers
        and the values are the number of candidates to generate for that iteration.
        If a list it must provide the number of candidates for each generator used. 
    """

    name = 'PoolParallelCollector'

    def __init__(self, num_candidates=None, **kwargs):        
        # Get the kwargs that are relevant for RayPoolUser
        RayPoolUser.__init__(self)
        # Give the rest to the CollectorBaseClass.
        CollectorBaseClass.__init__(self, **kwargs)        
        self.num_candidates = num_candidates

        self.generator_keys = []
        for generator in self.generators:
            key = self.pool_add_module(generator)
            self.generator_keys.append(key)
        
    def make_candidates(self):
        # We need to build the args and kwargs sent to the actors.
        number_of_candidates = self.get_number_of_candidates()
        
        # We need to determine which generators to use, these are passed 
        # as modules to the actors.
        # If sampler is empty and initialized, then use fallback generator
        if len(self.sampler) == 0 and self.sampler.initialized:
            fallback_generator = self.get_fallback_generator()
            generator_id = self.pool_add_module(fallback_generator)
            modules = [[generator_id]] * np.sum(number_of_candidates)
        else: 
            # This specifies which module each actor of the pool will use. 
            modules = []
            for generator_key, number in zip(self.generator_keys, number_of_candidates):
                modules += [[generator_key]] * number

        sampler_id = ray.put(self.sampler)
        environment_id = ray.put(self.environment)

        tasks = []
        for i in range(np.sum(number_of_candidates)):
            task = Task(function=remote_generate, 
                        modules=modules[i], 
                        args=[sampler_id, environment_id], 
                        kwargs={})
            tasks.append(task)

        candidates = self.task_map(tasks)

        # Flatten the output which is a list of lists. 
        flat_candidates = []
        for cand_list in candidates:
            for cand in cand_list:
                flat_candidates.append(cand.copy())
        return flat_candidates

    def get_number_of_candidates(self):
        if type(self.num_candidates) == list:
            return self.num_candidates
        elif type(self.num_candidates) == dict:
            return self.get_number_of_candidates_for_iteration()
            
    def get_number_of_candidates_for_iteration(self):
        # self.num_candidates must have this form: {0: [], 500: []}
        keys = list(self.num_candidates.keys())
        keys.sort()
        iteration = self.get_iteration_counter()
        if iteration is None:
            iteration = 0

        num_candidates = self.num_candidates[0] # yes, it must contain 0
        # now step through the sorted list (as long as the iteration is past the key) and extract the num_candidates
        # the last one extracted will be the most recent num_candidates enforced and should apply to this iteration
        for k in keys:
            if iteration < k:
                break
            num_candidates = self.num_candidates[k]
        return num_candidates





import pytest

from agox.utils.plot.utils import plane_to_indices


@pytest.mark.parametrize('plane,expected', [
    ('xy', (0, 1, 2, 1)),
    ('xy+', (0, 1, 2, 1)),
    ('xy-', (0, 1, 2, -1)),
    ('xz', (0, 2, 1, -1)),
    ('xz-', (0, 2, 1, 1)),
    ('yz+', (1, 2, 0, 1)),
    ('yx', (1, 0, 2, -1))
])
def test_plane_to_indices(plane, expected):
    indices = plane_to_indices(plane)
    assert indices == expected


@pytest.mark.parametrize('plane', [None, '', 'ab', 'abc', 'xx', 'xyz'])
def test_invalid_plane_to_indices(plane):
    with pytest.raises(ValueError, match='Invalid plane specification'):
        plane_to_indices(plane)



global mpl_backend
mpl_backend = 'Agg'

def get_mpl_backend():
    global mpl_backend
    return mpl_backend

def set_mpl_backend(new_backend):
    global mpl_backend
    mpl_backend = new_backend

def use_agox_mpl_backend():
    import matplotlib
    matplotlib.use(get_mpl_backend())


import numpy as np

from agox.utils.sparsifiers.ABC_sparsifier import SparsifierBaseClass


class RandomSparsifier(SparsifierBaseClass):
    name = "Random"

    def sparsify(self, X: np.ndarray = None) -> np.ndarray:
        if self.m_points < X.shape[0]:
            m_indices = np.random.choice(X.shape[0], size=self.m_points, replace=False)
            Xm = X[m_indices, :]

        else:
            m_indices = np.arange(0, X.shape[0])
            Xm = X

        return Xm, m_indices



from agox.generators.ABC_generator import GeneratorBaseClass
from agox.observer import Observer
import numpy as np

class ReuseGenerator(GeneratorBaseClass):

    """
    Reuses the candidates that were not selected for evaluation in the previous iteration.
    """


    name = 'ReuseGenerator'

    def __init__(self, order=8, gets={'candidates_key':'prioritized_candidates', 'evaluated_key':'evaluated_candidates'}, 
                 replace=True,**kwargs):
        super().__init__(order=order, gets=gets, replace=replace, **kwargs)
        self.prioritized_candidates = []
        self.evaluated_candidates = []

        self.add_observer_method(self.get_candidates_from_cache, sets={}, gets=self.gets[0], order=self.order[0])

    def _get_candidates(self, candidate, parents, environment):
        candidate = self.get_candidate_for_reuse()
        self.writer('ReuseCandidate', candidate)

        if candidate is None:
            return []

        return [candidate]

    def get_candidates_from_cache(self):        
        """
        Access the shared cache at the end of an iteration in order to get the most recently generated & prioritized 
        candidates and the most recently evaluated candidates. 
        """
        self.prioritized_candidates = self.get_from_cache(self.candidates_key)
        self.evaluated_candidates = self.get_from_cache(self.evaluated_key)

    def get_candidate_for_reuse(self):
        """
        Get the best candidate that was not selected for evaluation in the previous iteration. 

        This is not strictly neccesary if using an evaluator that pops candiadtes, but cannot guarantee that 
        and this not (too) slow anyway (currently). [Shouldnt scale poorly though].
        """
        if len(self.prioritized_candidates) == 0:
            return None
        for i, candidate in enumerate(self.prioritized_candidates):
            not_evaluated = True
            for eval_candidate in self.evaluated_candidates:
                if (candidate.positions == eval_candidate.positions).all():
                    self.prioritized_candidates.pop(i) # Remove the candidate.
                    not_evaluated = False
            if not_evaluated:
                return candidate
        return None
    
    def get_number_of_parents(self, sampler):
        return 0
    


import pytest
import numpy as np
from agox.generators import PermutationGenerator
from agox.test.generator_tests.generator_utils import generator_testing
from agox.test.test_utils import test_data_dicts

seed = 1
generator_args = []
generator_base_kwargs = {'c1':0.75, 'c2':1.25}
generator_class = PermutationGenerator

list_of_other_kwargs = [
    {}, # Tests that defaults havent changed. 
    {'max_number_of_swaps':1, 'rattle_strength':0.},
    {'max_number_of_swaps':3, 'rattle_strength':0.5},
    ]

# Because the PermutationGenerator only works for multicompoennt systems we filter abit:
from ase.io import read, write
multicomp_test_data_dicts = []
for test_data_dict in test_data_dicts:

    atoms = read(test_data_dict['path'])
    numbers = atoms.get_atomic_numbers()[len(atoms)-test_data_dict['remove']:]

    if len(np.unique(numbers)) >= 2:
        multicomp_test_data_dicts.append(test_data_dict)

for index, dictionary in enumerate(list_of_other_kwargs):
    list_of_other_kwargs[index] = (dictionary, index)

@pytest.fixture(params=list_of_other_kwargs)
def other_kwargs(request):
    return request.param

@pytest.mark.parametrize('test_data_dict', multicomp_test_data_dicts)
def test_generator(test_data_dict, other_kwargs, cmd_options):
    parameter_index = other_kwargs[1]
    other_kwargs = other_kwargs[0]

    create_mode = cmd_options['create_mode']
    test_mode = not create_mode
    tolerance = cmd_options['tolerance']

    generator_testing(generator_class, test_data_dict, generator_args, 
        generator_base_kwargs, other_kwargs, parameter_index, tolerance, 
        seed=seed, test_mode=test_mode)



from agox.generators.ABC_generator import GeneratorBaseClass
from ase.data import covalent_radii
from ase.io import write, Trajectory
import numpy as np
from scipy.spatial.distance import cdist
from ase.geometry import wrap_positions

from agox.generators.symmetry import SymmetryGenerator
from agox.generators.symmetry_utils.symmetry_enforcer import SymmetryEnforcer


class SymmetryRattleGenerator(SymmetryGenerator):

    """
    Rattles atoms in a seed structure while preserving symmetry or systematically 
    reducing symmetry.

    Parameters
    -----------
    n_rattle : int
        Medium number of atoms to rattle.
    rattle_amplitude : float
        Maximum distance to rattle atoms.
    let_lose_sym : float
        Probability of losing symmetry.
    let_switch : float
        Probability of switching points.
    write_strucs : bool
        If True, write structures to disk.
    verbose : bool
        If True, print verbose output.
    
    """

    name = "RattleGeneratorSym"

    def __init__(
        self,
        n_rattle=1,
        rattle_amplitude=3,
        let_lose_sym=0,
        let_switch=1.0,
        write_strucs=True,
        verbose=False,
        replace=True,
        **kwargs,
    ):
        super().__init__(replace=replace, **kwargs)
        self.n_rattle = n_rattle
        self.write_strucs = write_strucs
        self.rattle_amplitude = rattle_amplitude
        # probability of losing symmetry : float between 0 and 1 it is advised to set it considering the difficulty of the problem. The more diffuclt the lower
        self.let_lose_sym = let_lose_sym
        self.verbose = verbose
        # activates the switch points algorithm
        self.not_sym = []
        self.let_switch = let_switch
        self.move_on_point = True

    def _get_candidates(self, candidate, parents, environment):
        rang = 50
        already_lost = None
        keep_sym = True
        not_sym = []
        # candidate = sampler.get_random_member()
        # Happens if no candidates are part of the sample yet.
        if candidate is None:
            return []
        template = candidate.get_template()
        candinitial = candidate.copy()
        self.slab = template

        meta_required=["space_group", 
                        "table_mult", 
                        "non_sym",
                        "center"
        ]
        for meta in meta_required:
            candidate.add_meta_information(meta,parents[0].get_meta_information(meta)) 
            
        try:
            group_met = None
            # This is incorrect: always None
            group_met = candidate.get_meta_information("space_group")
            #should instead inherit meta information from parent 
            #group_met = parents[0].get_meta_information("space_group")

            if group_met is None:
                self.se = SymmetryEnforcer(self.slab)
                group_met = self.se.determine_plane_group(
                    candidate.positions[len(template) :]
                )

                num = candidate.numbers[len(template) :]
                table = []
                table_mult = []
                counter = 0
                num_ = None
                for j, elem in enumerate(num):
                    if j == 0 or elem == num_:
                        counter += 1
                        num_ = elem
                        continue
                    else:
                        table.append(counter)
                        num_ = elem
                        counter = 1
                table.append(counter)
                for ta in table:
                    num_f = ta // group_met["number_max"]
                    table_mult += [group_met["number_max"]] * num_f
                    if ta % group_met["number_max"] != 0:
                        table_mult += [ta % group_met["number_max"]]
                already_lost = 0
                center = None
                group_met = group_met["name"]
                if group_met != "p1":
                    print("symmetry found!")
            else:
                table_mult = list(candidate.get_meta_information("table_mult"))
                not_sym = candidate.get_meta_information("non_sym")
                center = candidate.get_meta_information("center")


            if group_met.__contains__("p") == False:
                mode = "cluster"
                rang = 20
            else:
                mode = "slab"
            self.se = SymmetryEnforcer(
                self.slab,
                confinement_cell=self.confinement_cell,
                cell_corner=self.confinement_corner,
                mode=mode,
            )
            if len(template) == 0:
                self.se.cell = candidate.get_cell()
            self.se.cluster_center = center
            gr = group_met
            self.se.force_group = group_met
            group_met = self.se.select_group(mode=mode)
            self.not_sym = not_sym
            already_lost = candidate.get_meta_information("lost_sym")
        except:
            group_met = {
                "name": "p1",
                "system": ["rectangular", "hexagonal", "square", "oblique", "rhombic"],
                "number_max": 1,
                "axes": None,
                "rotations": None,
                "points": None,
            }
            table_mult = [1] * len(environment.get_numbers())
            not_sym = []
            mode = None
            already_lost = 0
        if already_lost is None:
            already_lost = 0
        let_lose_sym = self.let_lose_sym ** (already_lost + 1)
        group = group_met
        if let_lose_sym > np.random.random():
            keep_sym = False
            sumtoreach = sum(table_mult)
            table_multcop = table_mult.copy()
            if mode == "cluster":
                su = np.sum(table_mult)
                group, table_mult = self.se.select_group_desym_cluster(
                    group_met, table_mult
                )
                assert su == np.sum(table_mult), (
                    str(table_multcop) + " " + str(table_mult)
                )
            else:
                group, table_mult = self.se.select_group_desym(group_met, table_mult)
                if group["name"] == "p2cmmg" and group_met["name"] == "p4g":
                    candidate.positions[len(template) :] -= [
                        0,
                        -candidate.cell[1][1] / 2,
                        0,
                    ]
                    candidate.positions = wrap_positions(
                        candidate.positions, candidate.get_cell(), candidate.pbc
                    )
                    print("translated")
        self.se.group = group
        rattled = False
        if self.se.group is not None:
            if (
                self.let_switch >= np.random.random()
                and table_mult is not None
                and group["name"] not in ["C1", "p1"]
            ):
                candiswitch = candidate.copy()
                candidate, table_mult2 = self.se.swap_some_points(
                    candidate, environment, table_mult
                )
                if candidate is None:
                    candidate = candiswitch
                else:
                    # write('resym_s.vasp', candidate,  format='vasp')
                    rattled = True
                    table_mult = table_mult2
        else:
            self.se.group = {
                "name": "p1",
                "system": ["rectangular", "hexagonal", "square", "oblique", "rhombic"],
                "number_max": 1,
                "axes": None,
                "rotations": None,
                "points": None,
            }
        indices_to_rattle = self.get_indices_to_rattle(candidate, table_mult, keep_sym)
        # write('resym'+str(self.se.group['name'])+'.vasp', candidate,  format='vasp')
        if self.verbose:
            print(
                group["name"]
                + " "
                + str(indices_to_rattle)
                + " "
                + str(let_lose_sym)
                + " "
            )
        for i in indices_to_rattle:
            i = int(i)
            vector_displac = None
            self.se.atomic_number = candidate.numbers[i]
            for _ in range(rang):  # For atom i attempt to rattle up to 30 times.
                candicopy = candidate.copy()
                if (
                    group is None
                    or group["number_max"] == 1
                    or candidate[i].number in not_sym
                    or already_lost > 1
                ):
                    # it is recommended to rattle gently symmetryc structures but no longer relevant for P1
                    radius = 2.5 * np.random.rand() ** (1 / self.dimensionality)
                else:
                    radius = self.rattle_amplitude * np.random.rand() ** (
                        1 / self.dimensionality
                    )
                displacement = self.get_displacement_vector(radius)
                posit = candidate.get_scaled_positions()
                if (
                    group is None
                    or group["number_max"] == 1
                    or candidate[i].number in not_sym
                ):
                    if not self.check_confinement(candidate[i].position + displacement):
                        continue
                    if self.check_new_position(
                        candidate,
                        candidate[i].position + displacement,
                        candidate[i].number,
                        skipped_indices=[i],
                    ):
                        candidate[i].position += displacement
                        rattled = True
                        break
                    continue
                if mode == "cluster":
                    vector_displac = self.se.identify_mult_point_cluster_new(
                        candidate, table_mult, i, displacement, keep_sym
                    )
                    if (
                        vector_displac is not False
                        and np.linalg.norm(vector_displac) > 1e-6
                    ):
                        displacement = (
                            np.array(vector_displac)
                            * ((-1) ** np.random.randint(1, 6))
                            * np.linalg.norm(displacement)
                        )
                    if vector_displac is False:
                        break
                else:
                    if vector_displac is None:
                        vector_displac = self.se.identify_mult_point_fast(
                            candidate, table_mult, i, displacement, keep_sym
                        )
                    if (
                        vector_displac is not False
                        and np.linalg.norm(vector_displac) > 1e-6
                    ):
                        vector_displac[2] = displacement[2] / np.linalg.norm(
                            displacement
                        )
                        displacement = (
                            np.array(vector_displac)
                            * ((-1) ** np.random.randint(1, 6))
                            * np.linalg.norm(displacement)
                        )
                    if vector_displac is False:
                        (
                            sug_pos,
                            vector_displac,
                            scal_pos,
                            vector_scaled,
                        ) = self.se.replace_on_ax(candidate.positions[i])
                        candidate[i].position = sug_pos[0]
                        if self.verbose:
                            print(
                                "unidentifed axis leads to "
                                + str(group["name"])
                                + " "
                                + str(posit[i])
                                + " ->"
                                + str(scal_pos)
                                + " with vect "
                                + str(vector_scaled)
                            )
                        displacement = (
                            np.array(vector_displac)
                            * ((-1) ** np.random.randint(1, 6))
                            * np.linalg.norm(displacement)
                        )
                sug_pos = [np.copy(candidate.positions[i]) + displacement]
                sug_pos = wrap_positions(sug_pos, candidate.cell, candidate.pbc)
                sug_pos = sug_pos[0]
                if not self.check_confinement(sug_pos).all():
                    continue
                candicopy[i].position = sug_pos
                # This part handles the merging algorithm and is probably the most not stable not well coded part of the rattle
                ind = self.se.get_position_in_table(candidate, i, table_mult)

                if self.check_new_position(
                    candidate,
                    sug_pos,
                    candidate[i].number,
                    skipped_indices=list(range(i, i + table_mult[ind])),
                    mic=False,
                    rmax=False,
                ):
                    merge, table_mult_new, new_t = self.se.merger(
                        candicopy, i, group, table_mult, cutoff=0.6, disp=displacement
                    )

                    if merge is None:
                        continue
                    # print(str(i)+ ' ' + ' '+str(resym.shape[0])+ ' '+str(table_mult)+' '+str(table_mult_new))
                    if len(merge) == len(candicopy.positions[i : i + merge.shape[0]]):
                        candicopy.positions[i : i + merge.shape[0]] = merge
                    else:
                        continue
                    if self.se.system == "cluster":
                        mic = False
                    else:
                        mic = True
                    if mode == "cluster" and self.check_new_position(
                        candicopy, mic=mic, rmax=False
                    ):
                        candidate.positions = candicopy.positions
                        table_mult = table_mult_new
                        assert np.sum(table_mult) == len(candidate)
                        rattled = True
                        break
                    elif mode != "cluster" and self.check_new_position(
                        candicopy, mic=mic, rmax=False
                    ):
                        candidate.positions[i] = sug_pos
                        ret_t = 0
                        # dirty implementation yet to improve
                        if len(new_t) > 1:
                            table_mult = table_mult_new
                            candidate.positions[i + new_t[0]] = merge[new_t[0]]
                        if len(new_t) > 2:
                            candidate.positions[i + new_t[0] + new_t[1]] = merge[
                                new_t[0] + new_t[1]
                            ]
                        if len(new_t) > 3:
                            candidate.positions[
                                i + new_t[0] + new_t[1] + new_t[2]
                            ] = merge[new_t[0] + new_t[1] + new_t[2]]
                        posit2 = candidate.get_scaled_positions()
                        """
                             try:
                                resym, table_mult_2=self.se.resymmetrize(candidate, group, table_mult, cutoff=0.5,keep_sym=keep_sym,indices_to_rattle=indices_to_rattle,not_sym=not_sym)
                             except:
                                continue
                             if resym is None or resym.shape[0]!=len( list(environment.get_numbers())) :
                                continue
                             candidate.positions[len(template):] =  resym
                             table_mult=table_mult_2
                             """
                        rattled = True
                        break
        if rattled:
            # if mode!='cluster' or keep_sym==False:
            if self.se.group["name"] not in ["C1", "p1"] and len(table_mult) > 1:
                copytabl = table_mult.copy()
                try:
                    resym, table_mult = self.se.resymmetrize(
                        candidate,
                        group,
                        table_mult,
                        cutoff=0.5,
                        keep_sym=keep_sym,
                        indices_to_rattle=indices_to_rattle,
                        not_sym=not_sym,
                    )
                except:
                    raise
                    return []
                if (
                    resym is None
                    or np.sum(table_mult) != (len(candidate) - len(template))
                    or resym.shape[0] != len(list(environment.get_numbers()))
                ):
                    print(
                        str(table_mult)
                        + " "
                        + str(self.se.group["name"])
                        + " "
                        + str(copytabl)
                    )
                    return []
                candidate.positions[len(template) :] = resym
            candidate.add_meta_information("space_group", group["name"])
            candidate.add_meta_information("table_mult", np.array(table_mult))
            if keep_sym == False and (group["number_max"] < 8 or mode != "cluster"):
                already_lost += 1
            candidate.add_write_key("space_group")
            return [candidate]
        else:
            return []

    def get_indices_to_rattle(self, candidate, table_mult=None, keep_sym=True):
        # Establish indices_to_rattle as the indices of the atoms that should be rattled
        n_rattle = self.n_rattle
        if keep_sym:
            if self.n_rattle < 2 and len(table_mult) >= 12:
                n_rattle = 2
            if self.n_rattle < 3 and len(table_mult) > 20 or max(table_mult) == 1:
                n_rattle = 6
        template = candidate.get_template()
        n_template = len(template)
        n_total = len(candidate)
        n_non_template = n_total - n_template
        probability = n_rattle / n_non_template
        indices_to_rattle = np.arange(n_template, n_total)[
            np.random.rand(n_non_template) < probability
        ]
        if len(self.not_sym) > 0:
            n_rattle += len(self.not_sym)
        if table_mult is not None:
            if max(table_mult) > 1:
                scale = (
                    abs(((len(table_mult) * n_rattle - n_rattle**2) / len(table_mult)))
                    ** 0.5
                )
                size = min(
                    len(table_mult), int(np.random.normal(loc=n_rattle, scale=scale))
                )
                if size < 1:
                    size = 1
                table = np.concatenate(([0], table_mult[:-1]))
                for i in range(len(table)):
                    if i > 0:
                        table[i] = table[i] + table[i - 1]
                table += len(template)
                indices_to_pick = table
                prob = np.array(table_mult) / n_non_template
                if np.sum(prob) != 1:
                    prob /= np.sum(prob)
                indices_to_rattle = np.random.choice(
                    indices_to_pick, p=prob, replace=False, size=size
                )
                np.sort(indices_to_rattle)

        if len(indices_to_rattle) == 0:
            indices_to_rattle = [np.random.randint(n_template, n_total)]
        # now indices_to_rattle is ready to use
        return indices_to_rattle

    def get_number_of_parents(self, sampler):
        return 1



import numpy as np
from agox.writer import agox_writer
from agox.samplers.ABC_sampler import SamplerBaseClass
from agox.observer import Observer

from typing import List
from agox.candidates import CandidateBaseClass as Candidate


class RateTracker:
    def __init__(self, sample_size: int):
        self.sample_size = sample_size
        self.acceptance = {i: list() for i in range(sample_size)}
        self.swap_up = {i: list() for i in range(sample_size)}
        self.swap_down = {i: list() for i in range(sample_size)}

    def update_acceptance(self, sample_index: int, value: int) -> None:
        self.acceptance[sample_index].append(value)

    def update_swap_up(self, sample_index: int, value: int) -> None:
        if sample_index < self.sample_size - 1:
            self.swap_up[sample_index].append(value)

    def update_swap_down(self, sample_index: int, value: int) -> None:
        if sample_index < self.sample_size:
            self.swap_down[sample_index].append(value)

    def get_acceptance_rate(
        self, sample_index: int, start=None, stop=None, step=None
    ) -> float:
        index_slice = slice(start, stop, step)
        return np.mean(self.acceptance[sample_index][index_slice])

    def get_swap_up_rate(
        self, sample_index: int, start=None, stop=None, step=None
    ) -> float:
        index_slice = slice(start, stop, step)
        return np.mean(self.swap_up[sample_index][index_slice])

    def get_swap_down_rate(
        self, sample_index: int, start=None, stop=None, step=None
    ) -> float:
        index_slice = slice(start, stop, step)
        return np.mean(self.swap_down[sample_index][index_slice])

    def get_swap_up(self, sample_index: int) -> bool:
        if len(self.swap_up[sample_index]) == 0:
            return False
        return bool(self.swap_up[sample_index][-1])

    def get_swap_down(self, sample_index: int) -> bool:
        if len(self.swap_down[sample_index]) == 0:
            return False
        return bool(self.swap_down[sample_index][-1])

    def get_acceptance(self, sample_index: int) -> bool:
        if len(self.acceptance[sample_index]) == 0:
            return False
        return bool(self.acceptance[sample_index][-1])


class ParallelTemperingSampler(SamplerBaseClass):
    name = "ParallelTempering"

    """
    Parallel tempering sampler for non-concurrent searches.

    Parameters
    ----------
    sample_size : int
        Number of members in the sample.
    swap : str
        Method for swapping members. Options are 'up' or 'down'.
    temperatures: List[float]
        List of temperatures to use.
    t_max : float
        Maximum temperature.
    t_min : float
    """

    def __init__(
        self,
        t_min=None,
        t_max=None,
        sample_size=5,
        swap="up",
        gets={"get_key": "candidates"},
        temperatures=None,
        swap_interval=10,
        always_accept=False,
        model=None,
        sets={"set_key": "candidates"},
        **kwargs,
    ):
        super().__init__(gets=gets, sets=sets, **kwargs)

        self.sample_size = sample_size
        self.swap_interval = swap_interval
        self.always_accept = always_accept
        self.model = model

        # Temperature logic:
        # Check that they are not all not none:
        if t_max is not None and t_min is not None and temperatures is not None:
            raise ValueError(
                "Either t_max and t_min or temperatures must be provided, not both."
            )
        elif (
            t_max is not None and t_min is not None
        ):  # If both are provided, generate the temperatures.
            self.temperatures = np.geomspace(t_min, t_max, sample_size)
        elif temperatures is not None:
            if len(temperatures) != sample_size:
                raise ValueError(
                    f"Length of temperatures ({len(temperatures)}) does not match sample size ({sample_size})."
                )
            else:
                self.temperatures = temperatures
        else:
            raise ValueError("Either t_max and t_min or temperatures must be provided.")

        self.reset_observer()  # We remove observers added by the base-class.
        self.add_observer_method(
            self.setup_sampler,
            gets=self.gets[0],
            sets=self.sets[0],
            order=self.order[0],
            handler_identifier="AGOX",
        )

        self.tracker = RateTracker(sample_size)

        if swap == "up":
            self.swap = self.swap_up
        elif swap == "down":
            self.swap = self.swap_down
        else:
            raise ValueError(
                f"Swap method '{swap}' not recognized. Use 'up' or 'down'."
            )

    @agox_writer
    @Observer.observer_method
    def setup_sampler(self, state):
        if self.do_check():
            evaluated_candidates = state.get_from_cache(self, self.get_key)
            evaluated_candidates = list(filter(None, evaluated_candidates))
            if len(evaluated_candidates) > 0:
                self.setup(evaluated_candidates)

        # Overwrite the candidates on the cache as the sampler may have updated meta information.
        state.add_to_cache(self, self.set_key, evaluated_candidates, mode="w")

    def setup(self, evaluated_candidates: List[Candidate]) -> None:
        if len(self.sample) != self.sample_size:
            self.setup_when_empty(evaluated_candidates)
            return

        self.update_sample(evaluated_candidates)

        iteration = self.get_iteration_counter()
        if self.decide_to_swap(iteration):
            self.swap()

        if iteration % 1 == 0:
            self.report_statistics()

    def update_sample(self, evaluated_candidates: List[Candidate]) -> None:
        for candidate in evaluated_candidates:
            sample_index = candidate.get_meta_information("walker_index")
            parent = self.sample[sample_index]
            temperature = self.temperatures[sample_index]

            # Determine if the candidate gets accepted:
            accepted = self.metropolis_check(candidate, parent, temperature)
            self.tracker.update_acceptance(sample_index, int(accepted))
            if accepted:
                self.sample[sample_index] = candidate

            candidate.add_meta_information("accepted", accepted)

    def metropolis_check(
        self, candidate: Candidate, parent: Candidate, temperature: float
    ) -> bool:
        if self.always_accept:
            return True

        E_candidate = self.get_potential_energy(candidate)
        E_parent = self.get_potential_energy(parent)

        if E_candidate < E_parent:
            accepted = True
        else:
            P = np.exp(-(E_candidate - E_parent) / temperature)
            accepted = P > np.random.rand()

        return accepted

    def swap_check(self, i, j):
        """
        Check if we should swap the i-th and j-th member of the sample.

        Parameters
        ----------
        i: int
            Index of the first member.
        j: int
            Index of the second member.

        Returns
        -------
        bool
            True if the swap is accepted, False otherwise.
        """
        E_i = self.get_potential_energy(self.sample[i]) 
        E_j = self.get_potential_energy(self.sample[j])
        beta_i = 1 / self.temperatures[i]
        beta_j = 1 / self.temperatures[j]

        P = np.min([1, np.exp((beta_i - beta_j) * (E_i - E_j))])

        return P > np.random.rand()

    def swap_up(self):
        self.writer("Swapping in 'up' mode")

        # Run over the sample starting from the bottom:
        for i in range(len(self.sample) - 1):
            swap_bool = self.swap_check(i, i + 1)

            self.tracker.update_swap_up(i, int(swap_bool))
            self.tracker.update_swap_down(i + 1, int(swap_bool))

            if swap_bool:
                self.sample[i], self.sample[i + 1] = self.sample[i + 1], self.sample[i]

    def swap_down(self):
        self.writer("Swapping in 'down' mode")

        # Run over the sample starting from the highest temperature:
        for i in range(len(self.sample) - 1, 0, -1):
            swap_bool = self.swap_check(i, i - 1)

            self.tracker.update_swap_up(i - 1, int(swap_bool))
            self.tracker.update_swap_down(i, int(swap_bool))

            if swap_bool:
                self.sample[i], self.sample[i - 1] = self.sample[i - 1], self.sample[i]

    def decide_to_swap(self, iteration_counter: int) -> bool:
        return iteration_counter % self.swap_interval == 0

    def setup_when_empty(self, evaluated_candidates: List[Candidate]) -> None:
        replace = len(evaluated_candidates) < self.sample_size
        indices = np.arange(len(evaluated_candidates))
        sample_indices = np.random.choice(
            indices, size=self.sample_size, replace=replace
        )
        self.sample = [evaluated_candidates[i] for i in sample_indices]

        if self.verbose:
            self.writer(
                "Initialized Metropolis Sampler with {}".format(len(self.sample))
            )

        for candidate in evaluated_candidates:
            candidate.add_meta_information("accepted", True)

    def report_statistics(self):
        self.writer(
            f'{"Member":^14s}|{"Energy":^10s}|{"Accept":^10s}|{"Up":^9s}|{"Down":^9s}|'
        )
        for i in range(self.sample_size):
            swap_up = self.get_swap_up_rate(i)
            swap_down = self.get_swap_down_rate(i)
            acceptance = self.get_acceptance_rate(i, start=-10)
            energy = self.sample[i].get_meta_information("pt_energy") or 0

            ati = "*" if self.tracker.get_acceptance(i) else ""
            iteration = self.get_iteration_counter()
            if self.decide_to_swap(iteration):
                suti = "*" if self.tracker.get_swap_up(i) else ""
                sdti = "*" if self.tracker.get_swap_down(i) else ""
            else:
                suti = ""
                sdti = ""

            try: 
                self.writer(
                    f"{i:^6d}({self.temperatures[i]:^02.3f}) |{energy:^10.3f}|{acceptance:^7.2f}{ati:^3s}|{swap_up:^6.2f}{suti:^3s}|{swap_down:^6.2f}{sdti:^3s}|"
                )
            except Exception as e:
                print(energy)
                print(e)

    def get_acceptance_rate(
        self, index: int, start=None, stop=None, step=None
    ) -> float:
        return self.tracker.get_acceptance_rate(index, start, stop, step)

    def get_swap_up_rate(self, index: int, start=None, stop=None, step=None) -> float:
        return self.tracker.get_swap_up_rate(index, start, stop, step)

    def get_swap_down_rate(self, index: int, start=None, stop=None, step=None) -> float:
        return self.tracker.get_swap_down_rate(index, start, stop, step)

    def get_potential_energy(self, candidate):
        if self.model is not None:
            E = self.model.predict_energy(candidate)
        elif hasattr(candidate, 'calc'):
            E = candidate.get_potential_energy()
        else:
            E = 0.0

        candidate.add_meta_information("pt_energy", float(E))
        return float(E)


from agox.generators.ABC_generator import GeneratorBaseClass
from ase.data import covalent_radii
from ase.geometry import get_distances
from ase.io import write
import numpy as np


class PermutationGenerator(GeneratorBaseClass):

    """
    Permutes atoms of different species in a seed structure.

    TODO: Clean-up this code. Its bad.

    Parameters
    -----------
    max_number_of_swaps : int
        Maximum number of permutations to perform.
    rattle_strength : float
        Maximum distance to permute atoms.
    use_xy_only : bool
        If True, only permute in xy-plane.
    ignore_H : bool
        If True, ignore hydrogen atoms.
    write_candidates_to_disk : bool
        If True, write candidates to disk.    

    """
    
    name='PermutationGenerator'

    def __init__(self, max_number_of_swaps=1, rattle_strength=0., 
        use_xy_only=False, ignore_H=False, write_candidates_to_disk=False, replace=True, 
        **kwargs):
        super().__init__(replace=replace,**kwargs)
        self.max_number_of_swaps = max_number_of_swaps
        self.rattle_strength = rattle_strength
        self.ignore_H = ignore_H
        self.use_xy_only = use_xy_only
        self.write_candidates_to_disk = write_candidates_to_disk

        if self.verbose:
            self.counter = -1

    def _get_candidates(self, candidate, parents, environment):
        if self.verbose:
            self.counter +=1
            
        if self.verbose and self.write_candidates_to_disk:
            write(f'candidate_{self.counter}.traj', candidate)
            
        template = candidate.get_template()

        number_of_atoms = len(candidate)
        number_of_template_atoms = len(candidate.get_template())
        number_of_non_template_atoms = number_of_atoms - number_of_template_atoms

        symbols = candidate.get_atomic_numbers()[number_of_template_atoms:]
        unique_symbols = np.unique(symbols)
        
        if self.ignore_H:
            unique_symbols = np.array([s for s in unique_symbols if s !=1])
            
        number_of_unique_symbols = len(unique_symbols)
        assert number_of_unique_symbols > 1, 'Cannot be used for single component systems'

        number_of_swaps = np.random.randint(self.max_number_of_swaps) + 1

        for n in range(number_of_swaps):
            symbol_i = unique_symbols[np.random.randint(number_of_unique_symbols)]
            remaining_symbols = np.delete(unique_symbols, [idx for idx in range(number_of_unique_symbols) if unique_symbols[idx]==symbol_i])

            symbol_j = remaining_symbols[np.random.randint(number_of_unique_symbols-1)]

            idx_symbol_i = np.argwhere(symbols==symbol_i).reshape(-1) + number_of_template_atoms
            idx_symbol_j = np.argwhere(symbols==symbol_j).reshape(-1) + number_of_template_atoms

            combinations_ij = np.array(np.meshgrid(idx_symbol_i, idx_symbol_j)).T.reshape(-1,2)

            for row in np.random.permutation(combinations_ij):
                swap_idx_i = row[0]
                swap_idx_j = row[1]
                
                new_positions = candidate.get_positions()
                new_positions[[swap_idx_i, swap_idx_j]] = new_positions[[swap_idx_j, swap_idx_i]]
                if self.use_xy_only:
                    new_positions[swap_idx_i] +=self.pos_add_disk(self.rattle_strength)
                    new_positions[swap_idx_j] +=self.pos_add_disk(self.rattle_strength)
                else:
                    new_positions[swap_idx_i] +=self.pos_add_sphere(self.rattle_strength)
                    new_positions[swap_idx_j] +=self.pos_add_sphere(self.rattle_strength)


                swap_successfull = True
                near_enough_to_other_atoms = False
                for i in (swap_idx_i, swap_idx_j):
                    for other_atom_idx in range(len(candidate)):
                        if other_atom_idx in [swap_idx_i, swap_idx_j]:
                            continue
                        other_atom = candidate[other_atom_idx]
                        
                        covalent_dist = covalent_radii[candidate[i].number] + covalent_radii[other_atom.number]
                        
                        rmin = 0.85 * covalent_dist
                        rmax = 1.15 * covalent_dist
                        tmp = np.linalg.norm(other_atom.position - new_positions[i])
                        if np.linalg.norm(other_atom.position - new_positions[i]) < rmin:
                            swap_successfull = False
                            break
                        if np.linalg.norm(other_atom.position - new_positions[i]) < rmax:
                            near_enough_to_other_atoms = True
                    if not swap_successfull or not near_enough_to_other_atoms:
                        break

                if not swap_successfull or not near_enough_to_other_atoms:
                    continue
                
                candidate.set_positions(new_positions)
                if self.verbose and self.write_candidates_to_disk:
                    write(f'candidate_swap_{n}_{self.counter}.traj', candidate)
                break
            else:
                self.writer('No swaps possible')
                return []
            
        return [candidate]

                    
    def pos_add_disk(self,rattle_strength):
        """Help function for rattling within a disk
        """
        r = rattle_strength * np.random.rand()**(1/2)
        theta = np.random.uniform(low=0, high=2*np.pi)
        pos_add = r * np.array([np.cos(theta),
                                np.sin(theta),
                                0])
        return pos_add

    def pos_add_sphere(self,rattle_strength):
        """Help function for rattling within a sphere
        """
        r = rattle_strength * np.random.rand()**(1/3)
        theta = np.random.uniform(low=0, high=2*np.pi)
        phi = np.random.uniform(low=0, high=np.pi)
        pos_add = r * np.array([np.cos(theta)*np.sin(phi),
                                np.sin(theta)*np.sin(phi),
                                np.cos(phi)])
        return pos_add        

    def get_number_of_parents(self, sampler):
        return 1



from abc import ABC, abstractmethod
import functools
from uuid import uuid4
from copy import copy

class Module:

    """
    Base class for all modules.

    Modules are the building blocks of the AGOX framework. They are used to define the
    different types of operations that are performed during a global optimization 
    algorithm.

    Parameters:
    -----------
    use_cache : bool
        If True, the module will use a cache to store the results of the methods.
    """
    kwargs = ['surname']

    def __init__(self, use_cache=False, surname=''):
        self.surname = surname
        
        self.use_cache = use_cache
        self.cache_key = str(uuid4())
        self.ray_key = str(uuid4())
        self.self_synchronizing = False

        self.dynamic_attributes = [] # Attributes that can be added and removed.
        self.tracked_attributes = [] # Attributes that the logger will track.
    
    def get_dynamic_attributes(self):
        return {key:self.__dict__.get(key, None) for key in self.dynamic_attributes}

    def add_dynamic_attribute(self, attribute_name):
        self.dynamic_attributes.append(attribute_name)

    def remove_dynamic_attribute(self, attribute_name):
        assert attribute_name in self.__dict__.keys()
        assert attribute_name in self.dynamic_attributes
        del self.dynamic_attributes[self.dynamic_attributes.index(attribute_name)]
    
    def get_tracked_attributes(self):
        attributes = dict()
        for key in self.tracked_attributes:
            try:
                attributes[key] = self.__dict__[key]
            except: # To avoid crashing if the attribute is not yet defined.
                attributes[key] = None
        return attributes
    
    def add_tracked_attribute(self, attribute_name):
        self.tracked_attributes.append(attribute_name)

    @property
    @abstractmethod
    def name(self):
        return NotImplementedError

    @property
    def dynamic_state(self):
        state = len(self.dynamic_attributes) > 0
        return state

    @property
    def __name__(self):
        """
        Defines the name.
        """
        if len(self.dynamic_attributes) > 0:
            last = 'Dynamic'
        else:
            last = ''
        return self.name + self.surname + last

    def find_submodules(self, in_key=None, only_dynamic=False):
        if in_key is None:
            in_key = []

        submodules = {}
        for key, value in self.__dict__.items():
            if issubclass(value.__class__, Module):
                key = in_key + [key]
                if only_dynamic:
                    if value.dynamic_state:                
                        submodules[tuple(key)] = value
                else:
                    submodules[tuple(key)] = value
                submodules.update(value.find_submodules(in_key=key, only_dynamic=only_dynamic))

        return submodules

    def set_for_submodule(self, submodule_keys, value):
        reference = self
        for key in submodule_keys[0:-1]:
            reference = self.__dict__[key]        
        reference.__dict__[submodule_keys[-1]] = value
        

    @classmethod
    def reset_cache_key(clc, func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            self.cache_key = str(uuid4())            
            return func(self, *args, **kwargs)
        return wrapper

def register_modules(object, modules, name):
    """
    Registers modules as attributes of an object. Used to let Tracker track 
    modules as attributes of modules that are only attributes of other modules.

    Parameters
    ----------
    object : object
        Object to register modules as attributes of.
    modules : list
        List of modules to register.
    name : str
        Name of the modules.

    Returns
    -------
    None.
    """
    for i, module in enumerate(modules):
        setattr(object, f'{name}_{i}', module)

    


from typing import Optional, Dict, Any, List

import numpy as np
from agox.candidates.standard import StandardCandidate
from agox.models.descriptors.spectral_graph_descriptor import \
    SpectralGraphDescriptor
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass


class DisjointFilteringPostprocess(PostprocessBaseClass):

    """
    Filters away disjont structures.
    
    Filter structures that are disjoint (i.e., have atoms in two or more
    separate groups, as determined by an adjacency graph). Internally, this
    postprocessor uses the `SpectralGraphDescriptor` class to evaluate
    eigenvalues of the Laplacian matrix.

    Parameters
    ----------
    graph_kwargs : Optional[Dict[str, Any]], optional
        Custom keyword arguments to pass to the `SpectralGraphDescriptor`
        class, by default None.
    """

    name = 'DisjointFilteringPostprocess'

    def __init__(self,
                 graph_kwargs: Optional[Dict[str, Any]] = None,
                 **kwargs):
        super().__init__(**kwargs)
        if graph_kwargs is None:
            graph_kwargs = {}

        self.descriptor = SpectralGraphDescriptor(environment=None,
                                                  mode='laplacian',
                                                  diagonal_mode='zero',
                                                  **graph_kwargs)

    def process_list(self, list_of_candidates: List[StandardCandidate]) -> List[Optional[StandardCandidate]]:
        """Process a list of candidates.

        Parameters
        ----------
        list_of_candidates : List[StandardCandidate]
            List of candidates.

        Returns
        -------
        List[Optional[StandardCandidate]]
            List of candidates, with disjoint candidates replaced by `None`.
        """
        processed_candidates = [self.postprocess(candidate) for candidate in list_of_candidates]

        filtered_count = processed_candidates.count(None)
        s = 's' if filtered_count != 1 else ''
        self.writer(f'Filtered {filtered_count} disjoint candidate{s}')

        return processed_candidates

    @PostprocessBaseClass.immunity_decorator
    def postprocess(self, candidate: StandardCandidate) -> Optional[StandardCandidate]:
        """Check if a candidate should be filtered.

        Parameters
        ----------
        candidate : StandardCandidate
            The candidate object to check.

        Returns
        -------
        Optional[StandardCandidate]
            The unchanged candidate object if its structure is non-disjoint,
            otherwise `None`.
        """
        # doi:10.1016/j.patcog.2008.03.011:
        # "The Laplacian has at least one zero eigenvalue, and the number of
        # such eigenvalues is equal to the number of disjoint parts in the
        # graph."
        eigenvalues = self.descriptor.create_features(candidate)
        disjoint = np.count_nonzero(np.isclose(eigenvalues, 0)) > 1

        if disjoint:
            return None
        else:
            return candidate



from agox.generators.ABC_generator import GeneratorBaseClass
from ase.data import covalent_radii
from ase.geometry import get_distances
from ase.io import write
import numpy as np


class ReplaceGenerator(GeneratorBaseClass):

    """
    Replaces atoms in a seed structure.

    Parameters
    -----------
    n_replace : int
        Medium number of atoms to replace.
    amplitude : float
        Maximum distance to replace atoms.
    replace : bool
        ?
        
    """

    name = 'ReplaceGenerator'

    def __init__(self, acquisitor=None, n_replace=5, amplitude=3, **kwargs):
        super().__init__(**kwargs)
        self.n_replace = n_replace
        self.amplitude = amplitude

    def _get_candidates(self, candidate, parents, environment):

        indices_to_move = self.get_indices_to_move(candidate)
        
        for index in list(indices_to_move):
            for _ in range(100):
                radius = self.get_radius(candidate)
                displacement = self.get_displacement_vector(radius)
                new_position = self.get_new_position_center(candidate, index, indices_to_move) + displacement

                # Check confinement limits:
                if not self.check_confinement(new_position).all():
                    continue

                if self.check_new_position(candidate, new_position, candidate[index].number, skipped_indices=indices_to_move):
                    candidate[index].position = new_position
                    indices_to_move.remove(index)
                    break

        return [candidate]

    def get_indices_to_move(self, candidate):
        number_of_atoms = len(candidate)
        template = candidate.get_template()
        number_of_template_atoms = len(template)
        number_of_non_template_atoms = number_of_atoms - number_of_template_atoms

        probability = self.n_replace/number_of_non_template_atoms
        randoms = np.random.rand(number_of_non_template_atoms)
        indices_to_move = np.arange(number_of_template_atoms,number_of_atoms)[randoms < probability]
        indices_to_move = np.random.permutation(indices_to_move)
        if len(indices_to_move) == 0:
            indices_to_move = [np.random.randint(number_of_template_atoms,number_of_atoms)]
        return list(indices_to_move)


    def get_new_position_center(self, candidate, index, indices_to_move):
        return candidate.positions[index]

    def get_radius(self, candidate):
        return self.amplitude * np.random.rand()**(1/self.dimensionality)
    
    def get_number_of_parents(self, sampler):
        return 1


import pytest
import numpy as np
from agox.generators import RattleGenerator
from agox.test.generator_tests.generator_utils import generator_testing
from agox.test.test_utils import test_data_dicts

seed = 1
generator_args = []
generator_base_kwargs = {'c1':0.75, 'c2':1.25}
generator_class = RattleGenerator

list_of_other_kwargs = [
    {},
    {'n_rattle':3, 'rattle_amplitude':3},
    {'n_rattle':5, 'rattle_amplitude':5}
    ]

for index, dictionary in enumerate(list_of_other_kwargs):
    list_of_other_kwargs[index] = (dictionary, index)

@pytest.fixture(params=list_of_other_kwargs)
def other_kwargs(request):
    return request.param

@pytest.mark.parametrize('test_data_dict', test_data_dicts)
def test_generator(test_data_dict, other_kwargs, cmd_options):
    parameter_index = other_kwargs[1]
    other_kwargs = other_kwargs[0]

    create_mode = cmd_options['create_mode']
    test_mode = not create_mode
    tolerance = cmd_options['tolerance']

    generator_testing(generator_class, test_data_dict, generator_args, 
        generator_base_kwargs, other_kwargs, parameter_index, tolerance, 
        seed=seed, test_mode=test_mode)



from typing import Optional, Tuple
import numpy as np
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import DotProduct
from agox.utils.sparsifiers.ABC_sparsifier import SparsifierBaseClass
from agox.models.GPR.priors import Repulsive
from agox.models.descriptors.type_descriptor import TypeDescriptor

class OnehotUniqueSparsifier(SparsifierBaseClass):

    name = 'OnehotUnique'

    def __init__(self, n_species):
        self.m_points = n_species

    def sparsify(self, X: np.ndarray) -> np.ndarray:
        return np.eye(self.m_points), None

class LocalMean(SparseGPR):

    name = 'LocalMeanModel'

    def __init__(self, species, database=None, prior=Repulsive(ratio=0.7), **kwargs):
        kernel = DotProduct(0)
        descriptor = TypeDescriptor.from_species(species)
        sparsifier = OnehotUniqueSparsifier(n_species=len(species))
        super().__init__(kernel=kernel, descriptor=descriptor, noise=0.001, use_ray=False,
                        n_optimize=0, prior=prior, sparsifier=sparsifier, 
                        database=database)

        self.dynamic_attributes = ["X", "Xm", "Kmm_inv", "alpha", "C_inv", 
                                   "kernel", "mean_energy"]


    def predict_forces(self, atoms, **kwargs):
        F = np.zeros((len(atoms), 3))
        return self._postprocess_forces(atoms, F)

    def train(self, data):
        super().train(data)

        self.writer('Energies for each species:')
        for alpha, species in zip(self.alpha.flatten(), self.descriptor.unique_numbers):
            self.writer(f'    Atomic number {species}: {alpha:.3f}')



from .ABC_sampler import SamplerBaseClass
import numpy as np
from ase.io import write

class KernelSimSampler(SamplerBaseClass):

    """
    Kernel Similarity Sampler for selecting candidates based on a kernel similarity criterion.

    Parameters
    ----------
    descriptor : agox.models.descriptor.ABC_descriptor.DescriptorBaseClass 
        Descriptor to use for the similarity criterion.
    model : agox.models.model.ABC_model.ModelBaseClass
        Model to use for the similarity criterion.
    log_scale : bool
        If True, the similarity criterion is scaled logarithmically.
    similarity_criterion : float
        The similarity criterion. If log_scale is True, this is the similarity criterion for the first sample.
    sample_size : int
        The number of samples to be taken.    
    """

    name = 'KernelSimSampler'

    def __init__(self, descriptor, model, kernel, log_scale=False,
        similarity_criterion=0.95, sample_size=10, **kwargs):
        super().__init__(**kwargs)
        self.descriptor = descriptor
        if log_scale:
            # 0.995 becomes: 0.995, 0.990, 0.980, 0.960, ...
            self.similarity_criteria = [1 - (1 - similarity_criterion) * 2**i for i in range(sample_size)]
        else:
            self.similarity_criteria = [similarity_criterion] * sample_size
        self.writer('SAMPLE_SIMILARITY_CRITERIA', self.similarity_criteria)

        self.sample_size = sample_size
        self.model = model
        self.kernel = kernel
        if self.kernel is None:
            self.kernel = model.kernel
        self.sample = []

    def setup(self, all_finished_structures):
        """
        Setup up the Kernel Similarity Sampler. 

        Parameters
        ----------
        all_finished_structures : list of ase.Atoms or agox.candidates.candidate.Candidate
            List of all finished structures.
        """

        if self.verbose:
            self.writer('len(all_finished_structures):',len(all_finished_structures))
        if len(all_finished_structures) == 0:
            return
        
        all_finished_structures.sort(key=lambda x: x.get_potential_energy())

        self.X = self.get_features(all_finished_structures)

        self.K0 = self.kernel(self.X[0].reshape(1, -1), self.X[0].reshape(1, -1))
        self.sample_indices = [0]
        for i, f_i in enumerate(self.X):
            this_is_a_new_structure = True
            for j,idx in enumerate(self.sample_indices):
                similarity = self.kernel(self.X[i].reshape(1, -1), self.X[idx].reshape(1, -1)) / self.K0
                if similarity > self.similarity_criteria[j]:
                    this_is_a_new_structure = False
                    break
            if this_is_a_new_structure:
                self.sample_indices.append(i)
                if len(self.sample_indices) >= self.sample_size:
                    break

        self.sample = [all_finished_structures[i] for i in self.sample_indices]
        if self.verbose:
            write('sample_{:06d}.traj'.format(self.get_iteration_counter()),self.sample)
        sample_energies = [t.get_potential_energy() for t in self.sample]
        self.writer('SAMPLE_DFT_ENERGY  ', '[',','.join(['{:8.3f}'.format(e) for e in sample_energies]),']')

        if self.model.ready_state:
            for s in self.sample:
                t = s.copy()
                t.set_calculator(self.model)
                E = t.get_potential_energy()
                sigma = self.model.get_property('uncertainty')
                s.add_meta_information('model_energy',E)
                s.add_meta_information('uncertainty',sigma)
            self.writer('SAMPLE_MODEL_ENERGY', \
                  '[',','.join(['{:8.3f}'.format(t.get_meta_information('model_energy')) for t in self.sample]),']')
            self.writer('SAMPLE_MODEL_SIGMA', \
                  '[',','.join(['{:8.3f}'.format(t.get_meta_information('uncertainty')) for t in self.sample]),']')

    def assign_to_best_sample(self,candidate_object):

        if len(self.sample) == 0:
            return None
        
        best_index = 0

        cluster_center = self.sample[best_index]
        cluster_center.add_meta_information('index_in_sampler_list_used_for_printing_purposes_only',best_index)

        return cluster_center

    def assign_to_closest_sample(self,candidate_object):

        if len(self.sample) == 0:
            return None
        
        # find out what cluster we belong to
        f_this = self.get_features([candidate_object])

        # assume that this is a new cluster that should be compared in energy to the highest one so far
        d_min_index = len(self.sample_indices) - 1

        # run through existing clusters in ascending energy order and stop first time we belong to one
        #   if no match, we already were assigned to the highest energy one
        for li,idx in enumerate(self.sample_indices):
            similarity = self.kernel.get_kernel(self.X[idx].reshape(1, -1),f_this) / self.K0
            if similarity > self.similarity_criteria[li]:
                d_min_index = li
                break

        cluster_center = self.sample[d_min_index]
        cluster_center.add_meta_information('index_in_sampler_list_used_for_printing_purposes_only',d_min_index)
        return cluster_center

    def adjust_sample_cluster_distance(self, chosen_candidate, closest_sample):
        # See if this structure had a lower energy than the lowest energy in the cluster it belongs to
        chosen_candidate_energy = chosen_candidate.get_potential_energy()
        closest_sample_energy = closest_sample.get_potential_energy()
        d_min_index = closest_sample.get_meta_information('index_in_sampler_list_used_for_printing_purposes_only')
        self.writer('CLUST_RES {:06d}'.format(self.get_iteration_counter()),
              '[',','.join(['{:8.3f}'.format(t.get_potential_energy()) for t in self.sample]),
              '] {:8.3f} {:8.3f}'.format(closest_sample_energy,chosen_candidate_energy),
              'NEW {:d} {}'.format(d_min_index,chosen_candidate.get_meta_information('description')) \
              if closest_sample_energy > chosen_candidate_energy else '')

    def get_features(self, structures):
        features = np.array(self.descriptor.get_features(structures)).sum(axis=1)
        return features



from agox.utils.metrics.metrics import get_metrics
from agox.utils.metrics.calibration import (
    calibration_curve,
    calibration_error,
    miscalibration_area,
    sharpness,
    dispersion,
)

__all__ = [
    "calibration_curve",
    "calibration_error",
    "miscalibration_area",
    "sharpness",
    "dispersion",
    "get_metrics",
]



import numpy as np
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
from sklearn.preprocessing import StandardScaler
from agox.observer import Observer
from agox.writer import agox_writer, Writer


class ScaleSelectDescriptor(DescriptorBaseClass, Observer, Writer):
    name = "ScaleSelectDescriptor"
    descriptor_type = "local"
    dynamic_attributes = ['scaler', 'selected_features']

    def __init__(
        self,
        base_descriptor,
        database=None,
        transfer_data=None,
        iteration_to_update=10,
        variance_threshold=0.1,
        scale=True,
        order=0,
        **kwargs
    ):
        DescriptorBaseClass.__init__(self, environment=base_descriptor.environment, use_cache=False, **kwargs)
        Observer.__init__(self, order=order)
        Writer.__init__(self, verbose=True, use_counter=True, prefix="")

        self.base_descriptor = base_descriptor
        self.descriptor_type = base_descriptor.descriptor_type
        self.variance_threshold = variance_threshold
        self.transfer_data = transfer_data if transfer_data is not None else []
        self.iteration_to_update = iteration_to_update
        self.scale = scale

        # Initialize for Ray.
        self.scaler = None
        self.selected_features = None

        if database is not None:
            self.add_observer_method(self.update_observer, sets=self.sets[0], gets=self.gets[0], order=self.order[0])
            self.attach(database)

    def update(self, dataset):
        self.scaler = StandardScaler()
        X = np.vstack([self.base_descriptor.get_features(atoms).reshape(len(atoms), -1) for atoms in dataset])
        self.scaler.fit(X)
        self.selected_features = self.scaler.scale_ > self.variance_threshold
        self.writer("Selected {} out of {} features".format(self.selected_features.sum(), len(self.selected_features)))

    def create_features(self, dataset):
        X = self.base_descriptor.get_features(dataset)
        if self.scaler is not None:
            if self.scale:
                X = self.scaler.transform(X)
            X = X[:, self.selected_features]
        return X

    def create_feature_gradient(self, dataset):
        X = self.base_descriptor.get_feature_gradient(dataset)
        if self.scaler is not None:
            if self.scale:
                X = X / self.scaler.scale_[None, None, None, :]
            X = X[:, :, :, self.selected_features]
        return X

    def get_number_of_centers(self, atoms):
        return self.base_descriptor.get_number_of_centers(atoms)

    @agox_writer
    @Observer.observer_method
    def update_observer(self, database, state):
        iteration = self.get_iteration_counter()
        if iteration >= self.iteration_to_update:
            data = database.get_all_candidates()
            all_data = data + self.transfer_data
            self.update(all_data)

    def __hash__(self):
        return 123456789 # EXTREMELY HACKY VERY DANGER?



import matplotlib

matplotlib.use("Agg")

import numpy as np
from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator, RattleGenerator
from agox.samplers import KMeansSampler
from agox.collectors import ParallelCollector
from agox.acquisitors import LowerConfidenceBoundAcquisitor
from agox.postprocessors import ParallelRelaxPostprocess
from ase import Atoms

from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR.kernels import RBF, Constant as C, Noise
from agox.models.GPR import GPR
from agox.models.GPR.priors import Repulsive

# Manually set seed and database-index
seed = 41
database_index = 0

##############################################################################
# Calculator
##############################################################################

from ase.calculators.vasp import Vasp

nslots = 8 # Number of cores
calc = Vasp(command='mpirun -np {} vasp'.format(nslots),
    prec="Low", lreal=False, lwave=False,
    xc="PBE", nwrite=2, istart=1, icharg=1,
    ispin=1, voskown=1, lmaxmix=6, nelm=500,
    nelmin=4, ibrion=-1, potim=0.3,
    iwavpr=1, nsw=0, ialgo=38, ismear=0,
    sigma=0.1, lorbit=11,
)

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au6",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=5)

##############################################################################
# Search Settings:
##############################################################################

# Setup a ML model.
descriptor = Fingerprint(environment=environment)
beta = 0.01
k0 = C(beta, (beta, beta)) * RBF()
k1 = C(1 - beta, (1 - beta, 1 - beta)) * RBF()
kernel = C(5000, (1, 1e5)) * (k0 + k1) + Noise(0.01, (0.01, 0.01))
model = GPR(descriptor=descriptor, kernel=kernel, database=database, prior=Repulsive())

# Sampler to choose candidates to modify using the generators.
sample_size = 10
sampler = KMeansSampler(
    descriptor=descriptor, database=database, sample_size=sample_size
)

# Generators to produce candidates structures
rattle_generator = RattleGenerator(**environment.get_confinement())
random_generator = RandomGenerator(**environment.get_confinement())

# Dict specificies how many candidates are created with and the dict-keys are iterations.
generators = [random_generator, rattle_generator]
num_candidates = {0: [10, 0], 5: [3, 7]}

# Collector creates a number of structures in each iteration. 
collector = ParallelCollector(
    generators=generators,
    sampler=sampler,
    environment=environment,
    num_candidates=num_candidates,
    order=1,
)

# Acquisitor to choose a candidate to evaluate in the real potential.
acquisitor = LowerConfidenceBoundAcquisitor(model=model, kappa=2, order=3)

# Number of steps is very low - should be set higher for a real search!
relaxer = ParallelRelaxPostprocess(
    model=acquisitor.get_acquisition_calculator(),
    constraints=environment.get_constraints(),
    optimizer_run_kwargs={"steps": 5},
    start_relax=8,
    order=2,
)


# Evaluator to evaluate the candidates in the real potential.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "prioritized_candidates"},
    optimizer_kwargs={"logfile": None},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 1},
    constraints=environment.get_constraints(),
    store_trajectory=True,
    order=4,
)

##############################################################################
# Let get the show running!
##############################################################################

# The oder of things here does not matter. But it can be simpler to understand
# what the expected behaviour is if they are put in order.
agox = AGOX(collector, relaxer, acquisitor, evaluator, database, seed=seed)

agox.run(N_iterations=10)



import numpy as np
from ase.data import covalent_radii
from abc import ABC, abstractmethod
from agox.candidates import StandardCandidate
from ase.geometry import get_distances                         
from agox.observer import Observer
from agox.writer import Writer, agox_writer
from uuid import uuid4
from typing import List

from agox.helpers.confinement import Confinement

dimensionality_angles = {
                        3:{'theta':[0, 2*np.pi], 'phi':[0, np.pi]},
                        2:{'theta':[0, 2*np.pi], 'phi':[np.pi/2, np.pi/2]},
                        1:{'theta':[0, 0], 'phi':[np.pi/2, np.pi/2]}
                        }

class GeneratorBaseClass(ABC, Observer, Writer, Confinement):

    def __init__(self, confinement_cell=None, confinement_corner=None, c1=0.75, c2=1.25, 
                use_mic=True, environment=None, sampler=None, gets={}, sets={'set_key':'candidates'}, order=2, 
                replace=True,verbose=True, use_counter=True, prefix='', surname=''):
        """
        use_mic: Whether to use minimum image convention when checking distances. 
                 If using a periodic cell but a smaller confinement cell set this to False to speed
                 up generation time! 
        sampler/enivronment: Needs to be set when not used with a collector!
        """
        Observer.__init__(self, sets=sets, gets=gets, order=order, surname=surname)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)
        Confinement.__init__(self, confinement_cell=confinement_cell, confinement_corner=confinement_corner)

        self.c1 = c1 # Lower limit on bond lengths. 
        self.c2 = c2 # Upper limit on bond lengths. 
        self.use_mic = use_mic
        self.environment = environment
        self.sampler = sampler
        self.replace = replace
        self.candidate_instanstiator = StandardCandidate

        if self.environment is not None:
            self.plot_confinement(environment)

        self.add_observer_method(self.generate,
                                 sets=self.sets[0], gets=self.gets[0], order=self.order[0],
                                 handler_identifier='AGOX')

    @abstractmethod
    def _get_candidates(self, candidate, parents, environment) -> List:
        """
        Method that returns a list of generated candidates
        """
        return []

    def get_candidates(self, sampler, environment):
        """
        Method used to get new candidates. 
        A generator method must have 'get_number_of_parents' method implemented.

        Parameters:
        -----------
        sampler: Sampler object
            Sampler object used to create and get sample members from.
        environment: Environment object
            The environment object used in the search.

        Returns: list
            List of candidates generated by the generator.
        """
        number_of_parents = self.get_number_of_parents(sampler)

        # If no parents are needed, the search template is given to generator.
        if number_of_parents == 0:
            parents = []
            candidate = environment.get_template().copy()
        # Checkt that the returned number of parents equal the required number.
        else:
            parents = sampler.get_parents(number_of_parents, replace=self.replace)
            if len(parents) != number_of_parents:
                return []
            candidate = parents[0].copy()
            candidate.reset_meta_information()
            candidate.set_constraint()

        # Generates new candidates and add relevant meta information.
        candidates = self._get_candidates(candidate, parents, environment)
        for candidate in candidates:
            candidate.add_meta_information('generator', self.name)
            candidate.add_meta_information('uuid', str(uuid4()))
            candidate.add_meta_information('parent_search_iterations', [p.get_meta_information('search_iterations') for p in parents])
            candidate.add_meta_information('parent_uuids', [p.get_meta_information('uuid') for p in parents])
        return candidates

    @property
    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def get_number_of_parents(self, sampler) -> int:
        """
        Method that returns the number of parents needed to generate a candidate.

        Parameters: 
        -----------
        sampler: Sampler object
            Sampler object used to create and get sample members from.
        
        Returns: Int
            Number of parents needed to generate a candidate.
        """
        pass

    def __call__(self, sampler, environment):
        return self.get_candidates(sampler, environment)
    
    ####################################################################################################################
    # Convenience methods:
    ####################################################################################################################

    def check_new_position(self, candidate, new_position, number, skipped_indices=[]):
        """
        Checks if new positions is not too close or too far to any other atom. 

        Probably not be the fastest implementation, so may be worth it to optimize at some point. 
        """
        state = True
        succesful = False
        for i in range(len(candidate)):
            if i in skipped_indices:
                continue

            covalent_dist_ij = covalent_radii[candidate[i].number] + covalent_radii[number]
            rmin = self.c1 * covalent_dist_ij
            rmax = self.c2 * covalent_dist_ij

            if self.use_mic:
                vector, distance = get_distances(new_position, candidate.positions[i], cell=candidate.cell, pbc=candidate.pbc)
            else:
                distance = np.linalg.norm(candidate.positions[i]-new_position)
            if distance < rmin: # If a position is too close we should just give up. 
                return False
            elif not distance > rmax: # If at least a single position is not too far we have a bond.
                succesful = True
        return succesful * state

    def get_sphere_vector(self, atomic_number_i, atomic_number_j):
        """
        Get a random vector on the sphere of appropriate radii. 

        Behaviour changes based on self.dimensionality: 
        3: Vector on sphere. 
        2: Vector on circle (in xy)
        1: Vector on line (x)
        """
        covalent_dist_ij = covalent_radii[atomic_number_i] + covalent_radii[atomic_number_j]
        rmin = self.c1 * covalent_dist_ij
        rmax = self.c2 * covalent_dist_ij
        r = np.random.uniform(rmin**self.dimensionality, rmax**self.dimensionality)**(1/self.dimensionality)
        return self.get_displacement_vector(r)

    def get_displacement_vector(self, radius):
        theta = np.random.uniform(*dimensionality_angles[self.dimensionality]['theta'])
        phi = np.random.uniform(*dimensionality_angles[self.dimensionality]['phi'])
        displacement = radius * np.array([np.cos(theta)*np.sin(phi),
                                          np.sin(theta)*np.sin(phi),
                                          np.cos(phi)])
        return displacement

    def get_box_vector(self):
        return self._get_box_vector(self.confinement_cell, self.confinement_corner) # From confinement-class.

    ####################################################################################################################
    # Observer functions
    ####################################################################################################################

    def start_candidate(self):
        """
        This method generates a candidate using the start generator, which allows other generators 
        to kick-start the sampler. 
        """
        from agox.generators import RandomGenerator
        return RandomGenerator(confinement_cell=self.confinement_cell, confinement_corner=self.confinement_corner, 
                c1=self.c1, c2=self.c2, use_mic=self.use_mic)(self.sampler, self.environment)

    @agox_writer
    @Observer.observer_method
    def generate(self, state):
        candidates = self.get_candidates(self.sampler, self.environment)
        if len(candidates) == 0 and self.sampler is not None and not self.sampler.initialized:
            candidates = self.start_candidate()
            self.writer('Fall-back to start generator, generated {} candidate '.format(len(candidates)))
        state.add_to_cache(self, self.set_key, candidates, mode='a')
        
    def plot_confinement(self, environment):
        if not self.confined:
            return

        import matplotlib.pyplot as plt
        from agox.utils.matplotlib_utils import use_agox_mpl_backend
        from agox.utils.plot import plot_atoms, plot_cell

        use_agox_mpl_backend()

        atoms = environment.get_template()
        atoms.set_constraint(environment.get_constraints())

        fig, axs = plt.subplots(1, 3, figsize=(15, 5))

        for ax, plane in zip(axs, ['xy', 'xz', 'yz']):
            plot_atoms(ax, atoms, plane=plane, plot_constraint=True)
            plot_cell(ax, atoms.cell, plane=plane,
                      collection_kwargs=dict(edgecolors='black', linestyles='dotted'))
            plot_cell(ax, self.confinement_cell, plane=plane, offset=self.confinement_corner,
                      collection_kwargs=dict(edgecolors='red', linestyles='dashed'))

        fig.savefig(f'confinement_plot_{self.name}.png')
        plt.close()



import pytest
from ase.io import read, write
from agox.environments import Environment
from agox.candidates import StandardCandidate
from ase import Atoms
import numpy as np

from agox.test.test_utils import test_data_dicts

@pytest.fixture(params=test_data_dicts)
def environment_and_dataset(request):
    atoms = read(request.param['path'])
    cell = atoms.get_cell()
    corner = np.array([0, 0, 0])
    remove = request.param['remove']
    numbers = atoms.get_atomic_numbers()[len(atoms)-remove:]

    template = read(request.param['path'])
    del template[len(template)-remove:len(template)]
    environment = Environment(template=template, numbers=numbers, confinement_cell=cell, 
            confinement_corner=corner)

    data = read(request.param['path'], ':')
    candidates = [StandardCandidate.from_atoms(template, a) for a in data]

    return environment, candidates

def test_environment(environment_and_dataset):
    environment, dataset = environment_and_dataset

    template = environment.get_template()
    assert template is not None
    environment.set_template(template)

    numbers = environment.get_numbers()
    assert isinstance(numbers, np.ndarray)
    environment.set_numbers(numbers)

    missing = environment.get_missing_types()
    assert isinstance(missing, np.ndarray)








import numpy as np
from ase.geometry import wrap_positions
from agox.helpers.confinement import Confinement
from ase.constraints import slice2enlist
from numpy.typing import NDArray
from typing import List
from ase import Atoms


class BoxConstraint(Confinement):
    def __init__(
        self,
        confinement_cell: NDArray = None,
        confinement_corner: NDArray = None,
        indices: NDArray = None,
        pbc: List[bool] = [False] * 3,
        wrap: bool = False,
        **kwargs,
    ):
        """
        Confinement constraint for atoms. This constraint will keep the atoms within a box defined by the cell and corner.

        Parameters
        ----------
        confinement_cell : np.ndarray
            The cell defining the box.
        confinement_corner : np.ndarray
            The corner of the box.
        indices : np.ndarray
            The indices of the atoms that are confined.
        pbc : list[bool]
            The periodic boundary conditions.
        wrap : bool
            If True, the atoms will be wrapped back into the box if they escape in directions
            with PBC.
        """

        super().__init__(
            confinement_cell=confinement_cell,
            confinement_corner=confinement_corner,
            indices=indices,
            pbc=pbc,
        )

        # Soft boundary & force decay.
        self.lower_soft_boundary = 0.05
        self.lower_hard_boundary = 0.001
        self.al, self.bl = np.polyfit(
            [self.lower_soft_boundary, self.lower_hard_boundary], [1, 0], 1
        )
        self.upper_soft_boundary = 0.95
        self.upper_hard_boundary = 0.999
        self.au, self.bu = np.polyfit(
            [self.upper_soft_boundary, self.upper_hard_boundary], [1, 0], 1
        )

        self.wrap = wrap

    @property
    def index(self):
        return self.indices

    @index.setter
    def index(self, value):
        self.indices = value

    def linear_boundary(self, x, a, b):
        return a * x + b

    def adjust_positions(self, atoms: Atoms, newpositions: NDArray):
        inside = self.check_confinement(newpositions[self.indices])
        # newpositions[not inside, :] = atoms.positions[not inside]
        # New positions of those atoms that are not inside (so outside) the box are set inside the box.

        if np.invert(inside).any():
            if self.wrap:
                newpositions[self.indices[np.invert(inside)], :] = wrap_positions(
                    newpositions[self.indices[np.invert(inside)], :],
                    cell=self.confinement_cell,
                    pbc=self.pbc,
                )
            for idx in self.indices[np.invert(inside)]:
                newpositions[idx, self.hard_boundaries] = atoms.positions[
                    idx, self.hard_boundaries
                ]

    def adjust_forces(self, atoms: Atoms, forces: NDArray):
        C = self.get_projection_coefficients(atoms.positions[self.indices])
        # Because adjust positions does not allow the atoms to escape the box we know that all atoms are witihn the box.
        # Want to set the forces to zero if atoms are close to the box, this happens if any component of C is close to 0 or 1.
        for coeff, idx in zip(C, self.indices):
            coeff = np.array([0.5 if p else c for c, p in zip(coeff, self.pbc)])
            if ((coeff < 0) * (coeff > 1)).any():
                forces[idx] = 0  # Somehow the atom is outside, so it is just locked.
            if (coeff > self.upper_soft_boundary).any():
                forces[idx, self.hard_boundaries] = (
                    self.linear_boundary(np.max(coeff), self.au, self.bu)
                    * forces[idx, self.hard_boundaries]
                )
            elif (coeff < self.lower_soft_boundary).any():
                forces[idx, self.hard_boundaries] = (
                    self.linear_boundary(np.min(coeff), self.al, self.bl)
                    * forces[idx, self.hard_boundaries]
                )

    def adjust_momenta(self, atoms: Atoms, momenta: NDArray):
        self.adjust_forces(atoms, momenta)

    def get_removed_dof(self, atoms: Atoms):
        return 0

    def index_shuffle(self, atoms: Atoms, ind: NDArray):
        # See docstring of superclass
        index = []

        # Resolve negative indices:
        actual_indices = set(np.arange(len(atoms))[self.index])

        for new, old in slice2enlist(ind, len(atoms)):
            if old in actual_indices:
                index.append(new)
        if len(index) == 0:
            raise IndexError("All indices in FixAtoms not part of slice")
        self.index = np.asarray(index, int)

    def get_indices(self):
        return self.index.copy()

    def todict(self):
        return {
            "name": "BoxConstraint",
            "kwargs": {
                "confinement_cell": self.confinement_cell.tolist(),
                "confinement_corner": self.confinement_corner.tolist(),
                "indices": self.indices.tolist(),
            },
        }


# To work with ASE read/write we need to do some jank shit.
from ase import constraints

constraints.__all__.append("BoxConstraint")
constraints.BoxConstraint = BoxConstraint

if __name__ == "__main__":
    from ase import Atoms
    from ase.io import write

    B = np.eye(3) * 1
    c = np.array([0, 0, 0])

    atoms = Atoms(
        "H3", positions=[[0.5, 0.5, 0.5], [0.1, 0.1, 0.9], [0.02, 0.9, 0.1]], cell=B
    )
    write("initial.traj", atoms)

    print(atoms.positions)
    BC = BoxConstraint(B, c, indices=np.array([0, 1, 2]), pbc=[False, False, True])
    atoms.set_constraint([BC])

    print(BC.check_if_inside_box(atoms.positions))
    forces = np.tile(np.array([[1, 1, 1]]), 3).reshape(3, 3).astype(float)
    print(forces)
    BC.adjust_forces(atoms, forces)
    print(forces)

    atoms.set_positions([[0.5, 0.5, 0.5], [0.1, 0.1, 0.9], [0.02, 1.1, -0.1]])
    print(atoms.positions)
    atoms.set_constraint([])
    write("moved.traj", atoms)

    # print(BC.upper_boundary_factor(0.999))
